"""Modules management."""

from collections import OrderedDict

from transaction import manager

from pyramid.config import Configurator

from ...lib.i18n import _
from ...lib.config import update_acl
from ...menu import MENU_ADMIN
from ...models import DB_METADATA, get_tm_dbsession
from .models import DBModule


PRINCIPALS_MODULES = (
    ('modules', _('Module management'), (
        ('viewer', _('View any module'), ('modules-view',)),
        ('editor', _('Edit or view any module'), (
            'modules-edit', 'modules-view'))
    )),
)
SUBMENU_MODULES = (
    None, _('Modules'), 'modules-view', 'modules_view', None)


# =============================================================================
def includeme(configurator):
    """Function to include modules functionality.

    :type  configurator: pyramid.config.Configurator
    :param configurator:
        Object used to do configuration declaration within the application.
    """
    # Populate/backup
    if not isinstance(configurator, Configurator):
        return

    # Routes
    configurator.add_route('modules_view', '/modules/view')
    configurator.add_route('modules_edit', '/modules/edit')

    # Premissions
    update_acl(configurator, PRINCIPALS_MODULES)
    if PRINCIPALS_MODULES[0] not in configurator.registry['principals']:
        configurator.registry['principals'].append(PRINCIPALS_MODULES[0])

    # Menu
    if 'modes' in configurator.registry:
        mode = dict(configurator.registry['modes']).get('admin')
        if mode is not None:
            mode[4][4].insert(0, SUBMENU_MODULES)
    elif 'menu' in configurator.registry and \
         MENU_ADMIN in configurator.registry['menu']:
        admin_menu = configurator.registry['menu'][
            configurator.registry['menu'].index(MENU_ADMIN)]
        admin_menu[4].insert(0, SUBMENU_MODULES)

    # Views
    configurator.scan('chrysalio.includes.modules.views')

    # Database table and Registry
    configurator.registry['modules'] = OrderedDict()
    configurator.registry['modules_off'] = set()
    with manager:
        dbsession = get_tm_dbsession(
            configurator.registry['dbsession_factory'], manager)
        DB_METADATA.create_all()
        for dbmodule in dbsession.query(DBModule):
            if dbmodule.inactive:
                configurator.registry['modules_off'].add(dbmodule.module_id)



# pylint: disable = import-outside-toplevel
"""Tests of ``modules.cioldap.views.cioldap.`` class."""

from json import dumps
from collections import namedtuple
from cgi import FieldStorage

from pyramid.httpexceptions import HTTPFound
from pyramid.response import Response

from ....tests import DBUnitTestCase


# =============================================================================
class UViewsCioLDAPCioLDAPView(DBUnitTestCase):
    """Unit test class for testing
    :class:`modules.cioldap.views.cioldap.CioLDAPView`."""

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        from ..models.dbldap import DBLdap, DBLdapProfile
        from ....models.dbprofile import DBProfile
        from ..lib.ldap import LDAP

        super(UViewsCioLDAPCioLDAPView, self).setUp()

        # pylint: disable = no-member
        dbsession = self.request.dbsession
        # pylint: enable = no-member
        dbldap = DBLdap(
            base='dc=iinov,dc=com',
            root_password='/ZSHiyU30mLjsr20KSXUoAD3tcragNnsQY+YxS2cIiA=',
            root_dn='cn=admin,dc=iinov,dc=com',
            user_dn='uid=_UID_,ou=people,dc=iinov,dc=com',
            user_filter='(&(objectclass=inetOrgPerson)(uid=_UID_))',
            field_first_name='givenName')
        dbldap.user_profiles.append(DBLdapProfile(profile_id='profile_viewer'))
        dbsession.add(dbldap)

        dbsession.add(DBProfile(
            profile_id='profile_creator',
            i18n_label=dumps({'en': 'Profile manager'})))
        dbsession.add(DBProfile(
            profile_id='profile_viewer',
            i18n_label=dumps({'en': 'Profile observer'})))
        dbsession.add(DBProfile(
            profile_id='user_viewer',
            i18n_label=dumps({'en': 'User account observer'})))

        self.configurator.add_route('home', '/')
        self.configurator.add_route('cioldap_edit', '/cioldap/edit')
        self.configurator.add_route('cioldap_view', '/cioldap/view')
        self.request.registry['authorities'] = {'ldap': LDAP()}

    # -------------------------------------------------------------------------
    def test_view(self):
        """[u:modules.cioldap.views.cioldap.CioLDAPView.view]"""
        from ..views.cioldap import CioLDAPView
        from ..models.dbldap import DBLdap
        from ....tests import DummyPOST

        # pylint: disable = no-member
        dbsession = self.request.dbsession
        # pylint: enable = no-member
        self.request.matched_route = namedtuple('Route', 'name')(
            name='cioldap_view')

        # With a LDAP configuration
        response = CioLDAPView(self.request).view()
        self.assertIn('form', response)
        self.assertIn('profile_labels', response)
        self.assertIn('dbldap', response)
        self.assertIsInstance(response['dbldap'], DBLdap)
        self.assertEqual(
            self.request.breadcrumbs.current_title(),
            'CioLDAP Module Configuration')

        # Without a LDAP configuration
        for dbldap in dbsession.query(DBLdap):
            dbsession.delete(dbldap)
        response = CioLDAPView(self.request).view()
        self.assertIsInstance(response, HTTPFound)

        # Ajax call
        self.request.is_xhr = True
        self.request.POST = DummyPOST()
        response = CioLDAPView(self.request).view()
        self.assertIsInstance(response, dict)
        self.assertFalse(response)

    # -------------------------------------------------------------------------
    def test_view_export(self):
        """[u:modules.cioldap.views.cioldap.CioLDAPView.view] export config"""
        from ..views.cioldap import CioLDAPView

        self.request.POST = {'exp!.x': True}

        response = CioLDAPView(self.request).view()
        self.assertIsInstance(response, Response)
        self.assertEqual(response.content_type, 'application/xml')

    # -------------------------------------------------------------------------
    def test_view_import(self):
        """[u:modules.cioldap.views.cioldap.CioLDAPView.view] import config"""
        from ....tests import DummyPOST
        from . import TEST1_LDAP_XML
        from ..views.cioldap import CioLDAPView
        from ..models.dbldap import LDAP_CHECK_INTERVAL, DBLdap

        self.request.POST = DummyPOST({'imp!.x': True})
        response = CioLDAPView(self.request).view()
        self.assertIn('dbldap', response)
        self.assertIsInstance(response['dbldap'], DBLdap)
        self.assertEqual(
            response['dbldap'].check_interval, LDAP_CHECK_INTERVAL)

        with open(TEST1_LDAP_XML, 'r') as hdl:
            input_file = FieldStorage()
            input_file.file = hdl
            input_file.filename = TEST1_LDAP_XML
            self.request.POST.multikeys = {'file': (input_file,)}
            response = CioLDAPView(self.request).view()
        self.assertEqual(response['dbldap'].check_interval, 0)

    # -------------------------------------------------------------------------
    def test_create(self):
        """[u:modules.cioldap.views.cioldap.CioLDAPView.edit] creation"""
        from ..views.cioldap import CioLDAPView
        from ..models.dbldap import DBLdap

        # pylint: disable = no-member
        dbsession = self.request.dbsession
        # pylint: enable = no-member
        self.request.matched_route = namedtuple('Route', 'name')(
            name='ldap_edit')
        for dbldap in dbsession.query(DBLdap):
            dbsession.delete(dbldap)

        response = CioLDAPView(self.request).edit()
        self.assertIn('form', response)
        self.assertIn('profile_labels', response)
        self.assertEqual(len(response['profile_labels']), 3)
        self.assertEqual(
            self.request.breadcrumbs.current_title(), 'CioLDAP Module Edition')

    # -------------------------------------------------------------------------
    def test_edit_import(self):
        """[u:modules.cioldap.views.cioldap.CioLDAPView.edit] import config"""
        from ....tests import DummyPOST
        from . import TEST1_LDAP_XML, TEST2_LDAP_XML
        from ..views.cioldap import CioLDAPView

        # A correct import
        self.request.POST = DummyPOST({'imp!.x': True})
        with open(TEST1_LDAP_XML, 'r') as hdl:
            input_file = FieldStorage()
            input_file.file = hdl
            input_file.filename = TEST1_LDAP_XML
            self.request.POST.multikeys = {'file': (input_file,)}
            response = CioLDAPView(self.request).edit()
        self.assertIsInstance(response, HTTPFound)

        # With missing base
        with open(TEST2_LDAP_XML, 'r') as hdl:
            input_file = FieldStorage()
            input_file.file = hdl
            input_file.filename = TEST2_LDAP_XML
            self.request.POST.multikeys = {'file': (input_file,)}
            response = CioLDAPView(self.request).edit()
        self.assertNotIsInstance(response, HTTPFound)

    # -------------------------------------------------------------------------
    def test_create_save(self):
        """[u:modules.cioldap.views.cioldap.CioLDAPView.edit] save creation"""
        from ..views.cioldap import CioLDAPView
        from ..models.dbldap import DBLdap

        # pylint: disable = no-member
        dbsession = self.request.dbsession
        # pylint: enable = no-member
        self.request.matched_route = namedtuple('Route', 'name')(
            name='ldap_edit')
        self.request.POST = {
            'host': 'localhost', 'port': '389', 'check_interval': '60',
            'user_dn': 'uid=_UID_,ou=people,dc=iinov,dc=com',
            'user_filter': '(&(objectclass=inetOrgPerson)(uid=_UID_))',
            'field_name': 'sn', 'pfl:profile_viewer': True,
            'sav!.x': True}
        for dbldap in dbsession.query(DBLdap):
            dbsession.delete(dbldap)

        # Missing base
        response = CioLDAPView(self.request).edit()
        errors = self.request.session.pop_flash('alert')
        self.assertEqual(len(errors), 1)
        self.assertIn('errors', errors[0])
        self.assertIn('form', response)

        # Correct form
        self.request.POST['base'] = 'cn=admin,dc=iinov,dc=com'
        CioLDAPView(self.request).edit()
        self.assertFalse(self.request.session.pop_flash('alert'))

    # -------------------------------------------------------------------------
    def test_edit(self):
        """[u:modules.cioldap.views.cioldap.CioLDAPView.edit]"""
        from ..views.cioldap import CioLDAPView
        from ..models.dbldap import DBLdap

        self.request.matched_route = namedtuple('Route', 'name')(
            name='ldap_edit')

        response = CioLDAPView(self.request).edit()
        self.assertIn('form', response)
        self.assertIn('profile_labels', response)
        self.assertIn('dbldap', response)
        self.assertIsInstance(response['dbldap'], DBLdap)
        self.assertEqual(
            self.request.breadcrumbs.current_title(), 'CioLDAP Module Edition')

    # -------------------------------------------------------------------------
    def test_edit_save(self):
        """[u:modules.cioldap.views.cioldap.CioLDAPView.edit] save"""
        from ..views.cioldap import CioLDAPView

        self.request.POST = {
            'host': 'localhost', 'port': '389', 'check_interval': '60',
            'base': 'cn=admin,dc=iinov,dc=com',
            'user_dn': 'uid=_UID_,ou=people,dc=iinov,dc=com',
            'user_filter': '(&(objectclass=inetOrgPerson)(uid=_UID_))',
            'field_name': 'sn', 'pfl:profile_viewer': False,
            'pfl:user_viewer': True, 'sav!.x': True}

        CioLDAPView(self.request).edit()
        self.assertIsInstance(CioLDAPView(self.request).edit(), HTTPFound)



"""Localization management."""

from sys import exit as sys_exit
from os.path import dirname, join
from locale import getdefaultlocale
from json import dumps, loads

import colander

from pyramid.i18n import TranslationStringFactory, make_localizer
from pyramid.exceptions import ConfigurationError

from .config import settings_get_list


_ = TranslationStringFactory('chrysalio')


# =============================================================================
def locale_negotiator(request):
    """Locale negotiator to figure out the language to use.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    :rtype: str
    """
    if request.session.get('lang'):
        return request.session['lang']
    return request.accept_language.lookup(
        settings_get_list(
            request.registry.settings, 'languages', ['en']),
        default_tag=request.registry['settings']['language'])


# =============================================================================
def add_translation_dirs(configurator, package):
    """Add one or more translation directory paths to the current configuration
    state according to settings and package name.

    :type  configurator: pyramid.config.Configurator
    :param configurator:
        Object used to do configuration declaration within the application.
    :param str package:
        Name of the calling package.
    """
    dirs = ['chrysalio:Locale', 'colander:locale']
    if package != 'chrysalio':
        dirs.insert(0, ('{0}:Locale'.format(package)))
    if configurator.get_settings().get('translation_dirs'):
        dirs = settings_get_list(
            configurator.get_settings(), 'translation_dirs') + dirs
    try:
        configurator.add_translation_dirs(*dirs)
    except (ImportError, ConfigurationError) as error:
        sys_exit('*** Translation directories: {0}'.format(error))


# =============================================================================
def translate(text, lang=None, request=None):
    """Return ``text`` translated.

    :param str text:
        Text to translate.
    :param str lang: (optional)
        Language to use.
    :type  request: pyramid.request.Request
    :param request: (optional)
        Current request to find the current language.
    :rtype: str
    """
    if lang is None and request is not None:
        lang = request.session.get('lang') \
            or request.locale_name or request.registry['settings']['language']
    return make_localizer(
        lang or getdefaultlocale()[0] or 'en',
        (join(dirname(__file__), '..', 'Locale'),)).translate(text)


# =============================================================================
def translate_field(request, i18n_fields, default=''):
    """Return the best translation according to user language.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    :param dict i18n_fields:
        Dictionary of avalaible translations.
    :param str default:
        Default label.
    :rtype: str
    """
    if not i18n_fields:
        return ''
    return \
        'lang' in request.session and i18n_fields.get(request.session['lang'])\
        or i18n_fields.get(request.locale_name) \
        or i18n_fields.get(request.registry['settings']['language']) \
        or i18n_fields.get(request.registry.settings.get(
            'pyramid.default_locale_name', 'en')) \
        or i18n_fields.get('en') or default


# =============================================================================
def record_format_i18n(record):
    """Update record by converting label and description entries into a
    dictionary with languages as key.

    :param dict record:
        Dictionary to process.
    :rtype: bool
    :return:
        ``True`` if a label has been found.
    """
    labels = loads(record['i18n_label']) if 'i18n_label' in record else {}
    descriptions = record.get('i18n_description', {})

    for k in tuple(record.keys()):
        if k.startswith('label_'):
            labels[k[6:]] = record[k]
            del record[k]
        elif k.startswith('description_'):
            descriptions[k[12:]] = record[k]
            del record[k]

    if not labels:
        return False

    record['i18n_label'] = dumps(labels, ensure_ascii=False)
    if descriptions:
        record['i18n_description'] = descriptions
    return True


# =============================================================================
def view_i18n_labels(
        request, form, dbitem, with_label=True, with_description=True):
    """Return HTML input to edit i18n labels.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    :param dbitem:
        Current SqlAlchemy object.
    :param bool with_label: (default=True)
       If ``True``, produce input fields for labels.
    :param bool with_description: (default=True)
       If ``True``, produce input fields for descriptions.
    :rtype: webhelpers2.html.literal
    """
    html = ''
    local_translate = request.localizer.translate

    if with_label and dbitem.i18n_label:
        labels = loads(dbitem.i18n_label)
        for lang in sorted(labels):
            html += form.grid_item(
                local_translate(_('Label (${l}):', {'l': lang})), labels[lang],
                clear=True)

    if with_description and dbitem.i18n_description:
        if dbitem.i18n_description:
            for lang in sorted(dbitem.i18n_description):
                html += form.grid_item(
                    local_translate(_('Description (${l}):', {'l': lang})),
                    dbitem.i18n_description[lang], clear=True)

    return html


# =============================================================================
def schema_i18n_labels(request, schema, label_len=0, description_len=0,
                       prefix='', required=True):
    """Update the Colander schema with fields for label ans, possibly, for
    descriptions of each available language.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    :type  schema: colander.SchemaNode
    :param schema:
       Current Colander schema.
    :param int label_len: (optional)
       Maximum length of label. If ``0``, no field is produced.
    :param int description_len: (optional)
       Maximum length of description. If ``0``, no field is produced.
    :param str prefix: (optional)
       A prefix for the name of field.
    :param bool required: (default=True)
       ``True`` if the label in default language is required.
    """
    default_lang = request.registry['settings']['language']

    # Label
    if label_len:
        for lang in sorted(settings_get_list(
                request.registry.settings, 'languages', ('en',))):
            if lang == default_lang and required:
                schema.add(colander.SchemaNode(
                    colander.String(),
                    name='{0}label_{1}'.format(prefix, lang),
                    validator=colander.Length(max=label_len)))
            else:
                schema.add(colander.SchemaNode(
                    colander.String(),
                    name='{0}label_{1}'.format(prefix, lang),
                    validator=colander.Length(max=label_len), missing=None))

    # Description
    if description_len:
        for lang in sorted(settings_get_list(
                request.registry.settings, 'languages', ('en',))):
            schema.add(colander.SchemaNode(
                colander.String(),
                name='{0}description_{1}'.format(prefix, lang),
                validator=colander.Length(max=description_len), missing=None))


# =============================================================================
def defaults_i18n_labels(
        dbitem, with_label=True, with_description=True, prefix=''):
    """Return a dictionary of default values for labels.

    :param dbitem:
        Current SqlAlchemy object.
    :param bool with_label: (default=True)
       If ``True``, produce input fields for labels.
    :param bool with_description: (default=True)
       If ``True``, produce input fields for descriptions.
    :param str prefix: (optional)
       A prefix for the name of field.
    """
    defaults = {}

    if with_label and dbitem.i18n_label:
        labels = loads(dbitem.i18n_label)
        for lang in labels:
            defaults['{0}label_{1}'.format(prefix, lang)] = labels[lang]

    if with_description and dbitem.i18n_description:
        for lang in dbitem.i18n_description:
            defaults['{0}description_{1}'.format(prefix, lang)] = \
                dbitem.i18n_description[lang]

    return defaults


# =============================================================================
def edit_i18n_labels(
        request, form, label_len=0, description_len=0, prefix=''):
    """Return HTML input to edit i18n labels.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    :param int label_len: (optional)
       Maximum length of label. If ``0``, no field is produced.
    :param int description_len: (optional)
       Maximum length of description. If ``0``, no field is produced.
    :param str prefix: (optional)
       A prefix for the name of field.
    :rtype: webhelpers2.html.literal
    """
    html = ''
    default_lang = request.registry['settings']['language']
    local_translate = request.localizer.translate

    # Label
    if label_len:
        for lang in sorted(settings_get_list(
                request.registry.settings, 'languages', ('en',))):
            html += form.grid_text(
                '{0}label_{1}'.format(prefix, lang),
                local_translate(_('Label (${l}):', {'l': lang})),
                required=lang == default_lang,
                maxlength=label_len, clear=True)

    # Description
    if description_len:
        for lang in sorted(settings_get_list(
                request.registry.settings, 'languages', ('en',))):
            html += form.grid_text(
                '{0}description_{1}'.format(prefix, lang),
                local_translate(_('Description (${l}):', {'l': lang})),
                maxlength=description_len, clear=True)

    return html



"""CioLDAP secutity functionalities."""

from ...lib.i18n import _
from ...lib.config import update_acl


PRINCIPALS_CIOLDAP = (
    ('cioldap', _('CioLDAP module management'), (
        ('viewer', _('View CioLDAP configuration'), ('cioldap-view',)),
        ('editor', _('Edit or view CioLDAP configuration'), (
            'cioldap-edit', 'cioldap-view'))
    )),
)


# =============================================================================
def includeme(configurator):
    """Function to include security.

    :type  configurator: :class:`pyramid.config.Configurator`
    :param configurator:
        Object used to do configuration declaration within the application.
    """
    update_acl(configurator, PRINCIPALS_CIOLDAP)



# pylint: disable = import-outside-toplevel
"""Tests of ``modules.cioldap.lib.ldap.LDAP`` class."""

from ....tests import DBUnitTestCase

from ..models.dbldap import DBLdap


# =============================================================================
class UViews(DBUnitTestCase):
    """Unit test class for testing
    :class:`modules.cioldap.lib.LDAP`."""

    # -------------------------------------------------------------------------
    def _add_ldap_settings(self):
        """Add LDAP settings."""
        from ..models.dbldap import DBLdapProfile

        # pylint: disable = no-member
        dbsession = self.request.dbsession
        # pylint: enable = no-member
        dbldap = DBLdap(
            host='ipa.demo1.freeipa.org',
            base='dc=demo1,dc=freeipa,dc=org',
            check_interval=0,
            user_dn='uid=_UID_,cn=users,cn=accounts,dc=demo1,'
            'dc=freeipa,dc=org',
            user_filter='(&(objectclass=person)(uid=_UID_))',
            field_first_name='givenName')
        dbsession.add(dbldap)
        dbsession.flush()
        dbldap.user_profiles.append(DBLdapProfile(profile_id='profile_viewer'))
        return dbldap

    # -------------------------------------------------------------------------
    def test_get(self):
        """[u:modules.cioldap.lib.ldap.LDAP.get]"""
        from json import dumps
        from webob.acceptparse import AcceptLanguageValidHeader
        from ....models.dbuser import DBUser
        from ....models.dbprofile import DBProfile
        from ..lib.ldap import LDAP

        self.request.accept_language = AcceptLanguageValidHeader(
            'fr-FR, fr;q=0.8, en-US;q=0.5, en;q=0.3')

        # Without password
        error = LDAP().get(self.request, 'user100', None, DBUser)[1]
        self.assertEqual('Password is mandatory.', error)

        # Without base
        error = LDAP().get(self.request, 'user100', 'user100pwd', DBUser)[1]
        self.assertEqual('LDAP configuration is missing.', error)

        # With an unknown user
        self._add_ldap_settings()
        ldap = LDAP()
        error = ldap.get(self.request, 'foo', 'Secret123', DBUser)[1]
        self.assertEqual('ID or password is incorrect.', error)

        # With a known user and unknown profile
        dbuser, error = ldap.get(self.request, 'admin', 'Secret123', DBUser)
        if error == 'ID or password is incorrect.':
            return
        self.assertIsNotNone(error)
        self.assertIn('without email', error)

        # With a known user and unknown profile
        dbuser, error = ldap.get(
            self.request, 'employee', 'Secret123', DBUser)
        if error != 'ID or password is incorrect.':
            self.assertIsInstance(dbuser, DBUser)
            self.assertEqual(dbuser.first_name, 'Test')
            self.assertEqual(dbuser.last_name, 'Employee')
            self.assertEqual(dbuser.email, 'employee@demo1.freeipa.org')
            self.assertFalse(dbuser.profiles)

        # With a known user and known profile
        # pylint: disable = no-member
        dbsession = self.request.dbsession
        # pylint: enable = no-member
        dbsession.add(DBProfile(
            profile_id='profile_viewer',
            i18n_label=dumps({'en': 'User profile viewer'})))
        dbuser, error = ldap.get(self.request, 'helpdesk', 'Secret123', DBUser)
        if error != 'ID or password is incorrect.':
            self.assertTrue(dbuser.profiles)

    # -------------------------------------------------------------------------
    def test_check(self):
        """[u:modules.cioldap.lib.ldap.LDAP.check]"""
        from ....lib.utils import encrypt
        from ....models.dbuser import DBUser
        from ..lib.ldap import LDAP

        dbuser = DBUser(
            login='test2', status='active', first_name='Guy',
            last_name='LIGUILI', honorific='Mr', email='test2@chrysal.io',
            authority='ldap')
        dbuser.set_password('test2pwd')

        # Without base
        error = LDAP().check(self.request, 'test2', 'employeepwd', dbuser)
        self.assertEqual('LDAP configuration is missing.', error)

        # With an unknown user
        dbldap = self._add_ldap_settings()
        ldap = LDAP()
        error = ldap.check(self.request, 'test2', 'test2pwd', dbuser)
        self.assertEqual('ID or password is incorrect.', error)

        # With a known user and an invalid password
        dbuser = DBUser(
            login='employee', status='active', first_name='Guy',
            last_name='LIGUILI', honorific='Mr', email='employee@chrysal.io',
            authority='ldap')
        dbuser.set_password('employeepwd')
        error = ldap.check(self.request, 'employee', 'employeepwd', dbuser)
        self.assertEqual('ID or password is incorrect.', error)

        # With a known user and a valid password
        error = ldap.check(self.request, 'employee', 'Secret123', dbuser)
        if error == 'ID or password is incorrect.':
            return
        self.assertIsNone(error)

        # Without password and an anonymous connection
        error = ldap.check(self.request, 'employee', None, dbuser)
        self.assertIsNone(error)

        # With cache
        dbldap.check_interval = 500
        ldap = LDAP()
        error = ldap.check(self.request, 'employee', None, dbuser)
        self.assertIsNone(error)
        error = ldap.check(self.request, 'employee', None, dbuser)
        self.assertIsNone(error)

        # Without password and with root_dn/root_password
        dbldap.check_interval = 0
        dbldap.root_dn = \
            'uid=admin,cn=users,cn=accounts,dc=demo1,dc=freeipa,dc=org'
        dbldap.root_password = encrypt('Secret123', 'ldap')
        error = LDAP().check(self.request, 'employee', None, dbuser)
        if error == 'ID or password is incorrect.':
            return
        self.assertIsNone(error)

        # With an invalid filter
        dbldap.base = 'dc=demo1,dc=freeipa2,dc=org'
        error = LDAP().check(self.request, 'employee', 'Secret123', dbuser)
        self.assertEqual(
            error, 'LDAP: unable to retrieve user information.')

        # With an invalid filter
        dbldap.user_filter = '(&(objectclass=group)(uid=_UID_))'
        error = LDAP().check(self.request, 'employee', 'Secret123', dbuser)
        self.assertIn('incorrect filter', error)



# pylint: disable = import-outside-toplevel
"""Tests of ``scripts.ciopopulate`` function and class."""

from os import makedirs, sep
from os.path import exists, join, dirname
from collections import namedtuple
from unittest import TestCase

from . import DBUnitTestCase


# =============================================================================
class UScriptsCIOPopulateMain(TestCase):
    """Unit test class for :func:`scripts.ciopopulate.main`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:scripts.ciopopulate.main]"""
        from ..scripts.ciopopulate import main
        from . import TEST_INI

        self.assertRaises(SystemExit, main)
        try:
            main([TEST_INI])
        except SystemExit:
            pass


# =============================================================================
class UScriptsCIOPopulatePopulate(DBUnitTestCase):
    """Unit test class for :class:`scripts.ciopopulate.Populate`."""

    # -------------------------------------------------------------------------
    def test_argument_parser(self):
        """[u:scripts.ciopopulate.Populate.argument_parser]"""
        from argparse import ArgumentParser
        from ..scripts.ciopopulate import Populate

        parser = Populate.argument_parser()
        self.assertIsInstance(parser, ArgumentParser)
        self.assertEqual(parser.description, 'Populate database.')

    # -------------------------------------------------------------------------
    def test_run(self):
        """[u:scripts.ciopopulate.Populate.run]"""
        from . import TEST_INI, TEST_DIR, TEST1_INI, TEST2_INI, TEST4_INI
        from . import TEST1_USR_XML, TEST3_USR_XML
        from ..scripts.ciopopulate import Populate
        from ..models.dbuser import DBUser
        from ..models.populate import xml2db
        from ..relaxng import RELAXNG

        # No database
        args = namedtuple(
            'ArgumentParser', 'conf_uri options lang attachments')(
                conf_uri=TEST1_INI, options=None, lang=None, attachments=None)
        self.assertEqual(Populate(args, DBUser, xml2db, RELAXNG).run(), 1)

        # No data
        args = namedtuple(
            'ArgumentParser', 'conf_uri options lang attachments')(
                conf_uri=TEST2_INI, options=None, lang=None, attachments=None)
        self.assertEqual(Populate(args, DBUser, xml2db, RELAXNG).run(), 1)

        # No attachments
        args = namedtuple(
            'ArgumentParser', 'conf_uri options lang attachments')(
                conf_uri=TEST4_INI, options=None, lang=None, attachments=None)
        self.assertEqual(Populate(args, DBUser, xml2db, RELAXNG).run(), 0)

        # Correct test.ini
        makedirs(join(TEST_DIR, 'Attachments'))
        args = namedtuple(
            'ArgumentParser', 'conf_uri options lang attachments')(
                conf_uri=TEST_INI, options=None, lang=None,
                attachments=join(
                    dirname(__file__), 'Attachments{0}'.format(sep)))
        self.assertEqual(Populate(args, DBUser, xml2db, RELAXNG).run(
            [TEST1_USR_XML, TEST3_USR_XML, TEST3_USR_XML]), 0)
        self.assertTrue(exists(
            join(TEST_DIR, 'Attachments', 'Users', 'Test1')))
        self.assertTrue(exists(
            join(TEST_DIR, 'Attachments', 'Users', 'Test5')))



"""Backup view callables."""

from os.path import exists, join, relpath
from time import time
from tempfile import mkdtemp
from shutil import rmtree
from zipfile import ZIP_DEFLATED, ZipFile, LargeZipFile
from configparser import ConfigParser

from lxml import etree
from sqlalchemy.exc import OperationalError
from sqlalchemy.orm.session import close_all_sessions
from transaction import manager

from pyramid.view import view_config
from pyramid.response import Response, FileResponse

from ..lib.i18n import _
from ..lib.utils import scandir, walk, tounicode
from ..lib.form import get_action, Form
from ..lib.xml import create_entire_xml
from ..lib.config import config_get_namespace
from ..lib.log import log_info, log_error
from ..models import DB_METADATA, get_tm_dbsession
from ..models.dbuser import DBUser
from ..models.populate import db2xml, xml2db, web2db
from . import BaseView


# =============================================================================
class BackupView(BaseView):
    """Class to manage backup and restore operations.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    """
    _DBUser = DBUser
    _db2xml = (db2xml,)
    _xml2db = (xml2db,)

    # -------------------------------------------------------------------------
    @view_config(
        route_name='backup', renderer='chrysalio:Templates/backup.pt',
        permission='backup-create')
    @view_config(
        route_name='backup', renderer='json', xhr=True,
        permission='backup-create')
    def index(self):
        """Choose between backup and restore."""
        # Ajax
        if self._request.is_xhr:
            self._restore()
            return {}

        # Action
        action = get_action(self._request)[0]
        if action == 'bck!':
            response = self._backup()
            if response is not None:
                return response
        if action == 'rst!':
            self._restore()

        # Breadcrumbs
        self._request.breadcrumbs(_('Configuration Backup'), 2)

        return {
            'form': Form(self._request), 'action': action,
            'download_max_size': self._request.registry['settings'][
                'download-max-size']}

    # -------------------------------------------------------------------------
    def _backup(self):
        """Backup the configuration into a ZIP or an XML file.

        :rtype: pyramid.response.Response
        """
        # Create the XML file
        elements = self._db2xml[0](
            self._request.dbsession,
            modules=self._request.registry.get('modules'))
        root_elt = create_entire_xml(
            self._request.registry['relaxng'], elements)
        # pylint: disable = protected-access
        if not isinstance(root_elt, etree._Element):  # pragma: nocover
            self._request.session.flash(root_elt, 'alert')
            return None
        # pylint: enable = protected-access

        # Return a XML file
        site_uid = self._request.registry.settings['site.uid']
        attachments = self._request.registry.settings.get('attachments')
        if not attachments or not exists(attachments) \
           or not tuple(scandir(attachments)):
            response = Response(
                body=etree.tostring(
                    root_elt, pretty_print=True, xml_declaration=True,
                    encoding='utf-8'),
                content_type='application/xml')
            response.headerlist.append((
                'Content-Disposition', 'attachment; filename="{0}.xml"'.format(
                    site_uid)))
            log_info(self._request, 'backup without attachments')
            return response

        # Return a Zip
        # pylint: disable = consider-using-with
        tmp_dir = mkdtemp(
            prefix=site_uid,
            dir=self._request.registry.settings.get('temporary'))
        zip_file = ZipFile(
            join(tmp_dir, '{0}.zip'.format(site_uid)), 'w', ZIP_DEFLATED)
        zip_file.writestr(
            '{0}.xml'.format(site_uid), etree.tostring(
                root_elt, encoding='utf-8', xml_declaration=True,
                pretty_print=True))
        for root, unused_, files in walk(attachments):
            for name in files:
                name = join(root, name)
                try:
                    zip_file.write(name, relpath(name, attachments))
                except LargeZipFile:  # pragma: nocover
                    zip_file.close()
                    rmtree(tmp_dir)
                    self._request.session.flash(
                        _('This configuration is too big!'), 'alert')
                    return None
                except IOError as error:  # pragma: nocover
                    zip_file.close()
                    rmtree(tmp_dir)
                    self._request.session.flash(error, 'alert')
                    return None
        zip_file.close()

        filename = zip_file.filename
        response = FileResponse(
            filename, request=self._request, content_type='application/zip')
        response.headers['Content-Disposition'] = \
            'attachment; filename="{0}.zip"'.format(site_uid)
        rmtree(tmp_dir)
        log_info(self._request, 'backup with attachments')
        return response

    # -------------------------------------------------------------------------
    def _restore(self):
        """Restore a configuration from a ZIP or an XML file.

        :rtype: bool
        """
        # Reset the old configuration
        reset = self._request.params.get('reset')
        admin = None
        if reset:
            admin = self._administrator_retrieve()
            if not self._drop_dbtables():  # pragma: nocover
                return False
            attachments = self._request.registry.settings.get('attachments')
            if attachments and exists(attachments):
                rmtree(attachments)

        # Restore the new configuration
        with manager:
            self._request.dbsession = get_tm_dbsession(
                self._request.registry['dbsession_factory'], manager)
            if not self._administrator_restore(admin):
                return False
            web2db(self._request, self._xml2db[0], error_if_exists=reset)
            if self._request.session.peek_flash('alert'):
                return False

        # To be continued
        self._request.session.flash(_('Backup restored.'))
        log_info(self._request, 'restore with reset' if reset else 'restore')
        if reset:
            self._request.session.clear()
            self._request.session['_creation_time'] = time()
            self._request.session['_accessed_time'] = time()
            self._request.response.delete_cookie(
                self._request.registry.settings.get('auth.cookie', 'CIO_AUTH'))
        return True

    # -------------------------------------------------------------------------
    def _administrator_retrieve(self):
        """Retrieve the administrator record.

        :rtype: dict
        """
        dbuser = self._request.dbsession.query(self._DBUser).filter_by(
            status='administrator').order_by('user_id').first()

        if dbuser is not None:
            return {
                k: dbuser.__dict__[k] for k in dbuser.__dict__
                if k[0] != '_' and k != 'user_id'}

        return None

    # -------------------------------------------------------------------------
    def _administrator_restore(self, record):
        """Restore the administrator record.

        :param dict record:
            Dictionary representing the administrator configuration.
        :rtype: bool
        """
        # Read administrator configuration from configuration file
        if record is None:
            config_file = self._request.registry.settings['__file__']
            config = ConfigParser({'here': config_file})
            config.read(tounicode(config_file), encoding='utf8')
            record = config_get_namespace(config, 'Populate', 'admin')

        error = self._DBUser.load_administrator(
            self._request.dbsession, record)
        if error:
            log_error(self._request, error)
            self._request.session.flash(error, 'alert')
            return False

        return True

    # -------------------------------------------------------------------------
    def _drop_dbtables(self):
        """Drop all tables of the database.

        :rtype: bool
        """
        # Drop
        try:
            close_all_sessions()
            DB_METADATA.drop_all()
        except OperationalError as error:  # pragma: nocover
            log_error(self._request, error.args[0])
            self._request.session.flash(error.args[0], 'alert')
            return False

        # Recreate
        try:
            DB_METADATA.create_all()
        except OperationalError as error:  # pragma: nocover
            log_error(self._request, error.args[0])
            self._request.session.flash(error.args[0], 'alert')
            return False

        return True



# pylint: disable = import-outside-toplevel
"""Tests of ``views.attachment`` functions."""

from . import DBUnitTestCase


# =============================================================================
class UViewsAttachmentAttchment(DBUnitTestCase):
    """Unit test class for testing :func:`views.attachment.attachment`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:views.attachment.attachment]"""
        from os.path import join, dirname
        from pyramid.response import FileResponse
        from pyramid.httpexceptions import HTTPNotFound
        from ..views.attachment import attachment

        self.assertRaises(HTTPNotFound, attachment, self.request)

        self.configurator.get_settings()['attachments'] = join(
            dirname(__file__), 'Attachments')
        self.request.matchdict = {'path': ('Users', 'Test1', 'test5.svg')}
        self.assertRaises(HTTPNotFound, attachment, self.request)

        self.request.matchdict = {'path': ('Users', 'Test5', 'test5.svg')}
        response = attachment(self.request)
        self.assertIsInstance(response, FileResponse)
        self.assertEqual(response.status_code, 200)


# =============================================================================
class UViewsAttachmentFavicon(DBUnitTestCase):
    """Unit test class for testing :func:`views.attachment.favicon`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:views.attachment.favicon]"""
        from os.path import join, dirname
        from pyramid.response import FileResponse
        from ..views.attachment import favicon

        response = favicon(self.request)
        self.assertIsInstance(response, FileResponse)
        self.assertEqual(response.status_code, 200)

        self.request.registry.settings['site.favicon'] = join(
            dirname(__file__), '..', 'Static', 'favicon.ico')
        response = favicon(self.request)
        self.assertIsInstance(response, FileResponse)
        self.assertEqual(response.status_code, 200)


# =============================================================================
class UViewsAttachmentRobots(DBUnitTestCase):
    """Unit test class for testing :func:`views.attachment.robots`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:views.attachment.robots]"""
        from os.path import join, dirname
        from pyramid.response import FileResponse
        from ..views.attachment import robots

        response = robots(self.request)
        self.assertIsInstance(response, FileResponse)
        self.assertEqual(response.status_code, 200)

        self.request.registry.settings['site.robots'] = join(
            dirname(__file__), '..', 'Static', 'robots.txt')
        response = robots(self.request)
        self.assertIsInstance(response, FileResponse)
        self.assertEqual(response.status_code, 200)



"""Error view callables."""

from pyramid.view import view_config
from pyramid.view import notfound_view_config, forbidden_view_config
from pyramid.httpexceptions import HTTPFound, HTTPBadRequest
from pyramid.security import NO_PERMISSION_REQUIRED

from ..lib.i18n import _
from ..lib.form import button


# =============================================================================
@forbidden_view_config(renderer='chrysalio:Templates/error.pt')
@notfound_view_config(renderer='chrysalio:Templates/error.pt')
@view_config(context=HTTPBadRequest, renderer='chrysalio:Templates/error.pt',
             permission=NO_PERMISSION_REQUIRED)
@forbidden_view_config(renderer='json', xhr=True)
@notfound_view_config(renderer='json', xhr=True)
@view_config(context=HTTPBadRequest, renderer='json', xhr=True)
def error_view(request):
    """This view outputs an error message or redirects to login page.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    """
    status = request.exception.status_int

    # AJAX
    if request.is_xhr:
        return {'error': status}

    # Not authenticated and 401 (unauthorized) or 403 (forbidden)
    if status in (401, 403) and request.authenticated_userid is None:
        return HTTPFound(
            request.route_path('login', _query=(('next', request.path),)))

    # Comment adjustment
    if status == 400 and \
       request.exception.title in ('Bad CSRF Token', 'Bad CSRF Origin'):
        request.exception.comment = _(
            'Access is denied: the server can not verify that your protection '
            'against cross-site request forgery (CSRF) is correct.')

    # 400 (bad request), 401 (unauthorized), 403 (forbidden), 404 (not found)
    request.response.status = status
    return {
        'title': _('Error ${status}', {'status': status}),
        'status': status,
        'message': request.exception.comment or {
            400: _('The server could not comply with the request since'
                   ' it is either malformed or otherwise incorrect.'),
            401: _('This server could not verify that you are authorized to'
                   ' access the document you requested.'),
            403: _('Access was denied to this resource.'),
            404: _('The resource could not be found.'),
            500: _('Internal Server Error.')}.get(
                status, request.exception.explanation),
        'button': button}



"""CioLDAP, a module to manage LDAP authentication."""

from pyramid.config import Configurator

from ...lib.i18n import _
from ...includes.modules.models import DBModule
from .. import Module
from .relaxng import RELAXNG_CIOLDAP
from .security import PRINCIPALS_CIOLDAP
from .lib.ldap import LDAP
from .models.populate import xml2db as _xml2db, db2xml as _db2xml


# =============================================================================
def includeme(configurator):
    """Function to register the module.

    :type  configurator: pyramid.config.Configurator
    :param configurator:
        Object used to do configuration declaration within the application.
    """
    # Registration
    Module.register(configurator, ModuleCioLDAP)
    if not isinstance(configurator, Configurator):
        return

    # Permission
    configurator.include('chrysalio.modules.cioldap.security')

    # Routes
    configurator.include('chrysalio.modules.cioldap.routes')

    # Views
    configurator.scan('chrysalio.modules.cioldap.views')


# =============================================================================
class ModuleCioLDAP(Module):
    """Class for CioLDAP module.

    :param str config_ini:
        Absolute path to the configuration file (e.g. development.ini).
    """

    name = _('LDAP')
    implements = ('ldap',)
    relaxng = RELAXNG_CIOLDAP
    xml2db = (_xml2db,)
    db2xml = (_db2xml,)
    areas = {}
    _DBModule = DBModule

    # -------------------------------------------------------------------------
    def activate(self, registry, dbsession):
        """Method to activate the module.

        :type  registry: pyramid.regisry.Registry
        :param registry:
            Application registry.
        :type  dbsession: sqlalchemy.orm.session.Session
        :param dbsession:
            SQLAlchemy session.
        """
        # Security
        if PRINCIPALS_CIOLDAP[0] not in registry['principals']:
            registry['principals'].append(PRINCIPALS_CIOLDAP[0])

        # Authorities
        if 'authorities' not in registry:
            registry['authorities'] = {}
        if 'ldap' not in registry['authorities']:
            registry['authorities']['ldap'] = LDAP()

    # -------------------------------------------------------------------------
    def deactivate(self, registry, dbsession):
        """Method to deactivate the module.

        :type  registry: pyramid.registry.Registry
        :param registry:
            Application registry.
        :type  dbsession: sqlalchemy.orm.session.Session
        :param dbsession:
            SQLAlchemy session.
        """
        # Security
        if PRINCIPALS_CIOLDAP[0] in registry['principals']:
            registry['principals'].remove(PRINCIPALS_CIOLDAP[0])

        # Authorities
        if 'authorities' in registry and 'ldap' in registry['authorities']:
            del registry['authorities']['ldap']

    # -------------------------------------------------------------------------
    def configuration_route(self, request):
        """Return the route to configure this module.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        """
        return request.route_path('cioldap_view')



# pylint: disable = import-outside-toplevel
"""Unit, integration, and functional testing."""

from os import makedirs
from os.path import join, dirname, exists, normpath
from collections import OrderedDict, namedtuple
from shutil import rmtree
from unittest import TestCase

from pkg_resources import register_loader_type, DefaultProvider
from _pytest.assertion.rewrite import AssertionRewritingHook
from transaction import manager
from webtest import TestApp as BaseTestApp

from pyramid.security import Authenticated, Allow
from pyramid.testing import setUp, tearDown, DummyRequest

from ..models.dbuser import DBUser


TEST_INI = join(dirname(__file__), 'test.ini')
TEST1_INI = join(dirname(__file__), 'test1.ini')
TEST2_INI = join(dirname(__file__), 'test2.ini')
TEST3_INI = join(dirname(__file__), 'test3.ini')
TEST4_INI = join(dirname(__file__), 'test4.ini')
TEST5_INI = join(dirname(__file__), 'test5.ini')
TEST6_INI = join(dirname(__file__), 'test6.ini')
TEST7_INI = join(dirname(__file__), 'test7.ini')
TEST8_INI = join(dirname(__file__), 'test8.ini')
CIOUPDATE_INI = join(dirname(__file__), 'cioupdate.ini')
CIOUPDATE1_INI = join(dirname(__file__), 'cioupdate1.ini')
CIOUPDATE2_INI = join(dirname(__file__), 'cioupdate2.ini')
CIOUPDATE3_INI = join(dirname(__file__), 'cioupdate3.ini')
CIOUPDATE4_INI = join(dirname(__file__), 'cioupdate4.ini')
CIOUPDATE5_INI = join(dirname(__file__), 'cioupdate5.ini')
CIOUPDATE6_INI = join(dirname(__file__), 'cioupdate6.ini')
TEST1_SET_XML = join(dirname(__file__), 'Xml', 'test1.cioset.xml')
TEST1_PFL_XML = join(dirname(__file__), 'Xml', 'test1.ciopfl.xml')
TEST1_USR_XML = join(dirname(__file__), 'Xml', 'test1.ciousr.xml')
TEST2_USR_XML = join(dirname(__file__), 'Xml', 'test2.ciousr.xml')
TEST3_USR_XML = join(dirname(__file__), 'Xml', 'test3.ciousr.xml')
TEST1_GROUP_XML = join(dirname(__file__), 'Xml', 'test1.ciogrp.xml')
I18N_XML = join(dirname(__file__), 'Xml', 'i18n.xml')
TEST_UTF8 = join(dirname(__file__), 'Texts', 'utf8.txt')
TEST_LATIN1 = join(dirname(__file__), 'Texts', 'latin1.rst')
TRANSFORM_XSL = join(dirname(__file__), 'Xml', 'transform.xsl')
TEST1_USR_ZIP = join(dirname(__file__), 'Backup', 'test1.ciousr.zip')
TEST2_USR_ZIP = join(dirname(__file__), 'Backup', 'test2.ciousr.zip')
TEST3_USR_ZIP = join(dirname(__file__), 'Backup', 'test3.ciousr.zip')
TEST_CSS = join(dirname(__file__), '..', 'Static', 'Css', 'chrysalio.css')
TEST_XXX = join(dirname(__file__), 'test.xxx')
EMAIL_HTML = join(dirname(__file__), 'Templates', 'email_test.pt')
EMAIL_TEXT = join(dirname(__file__), 'Templates', 'email_test.txt')
ATTACHMENTS_DIR = join(dirname(__file__), 'Attachments')
TEST1_SVG = join(ATTACHMENTS_DIR, 'Users', 'Test1', 'test1.svg')
TEST5_SVG = join(ATTACHMENTS_DIR, 'Users', 'Test5', 'test5.svg')

TEST_DIR = normpath(join(dirname(__file__), '..', '..', 'Test~'))
BACKUP_DIR = join(TEST_DIR, 'Backup')

FTP_HOST = 'ftp.chrysal.io'
FTP_USER = 'ftp-chrysalio'
FTP_PASSWORD = 'hFphq3c6Qj9TgaB1X1mtYVj9q9ytTdIIeU6rlyxAb+I='
RESTFUL_HOST = 'https://demo.chrysal.io/'
RESTFUL_LOGIN = 'user1'
RESTFUL_KEY = 'Cb4rCDyPJZbkuBNVFMLAtxgzJ4QJX/qt6RHyPRy3g4I='


# =============================================================================
class ConfiguratorTestCase(TestCase):
    """Base class for unit test of a module."""

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        self.configurator = setUp()

    # -------------------------------------------------------------------------
    def tearDown(self):
        """Undo the effects of ``pyramid.testing.setUp()``."""
        tearDown()


# =============================================================================
class ModuleTestCase(TestCase):
    """Base class for unit test of a module."""

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        self.configurator = setUp()
        self.configurator.set_root_factory(DummyRootFactory)
        self.configurator.registry['modules'] = OrderedDict()
        self.configurator.registry['modules_off'] = set()

    # -------------------------------------------------------------------------
    def tearDown(self):
        """Undo the effects of ``pyramid.testing.setUp()``."""
        tearDown()


# =============================================================================
class TmpDirTestCase(TestCase):
    """Base class for unit test with a temprary directory."""

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        if exists(TEST_DIR):
            rmtree(TEST_DIR)
        makedirs(TEST_DIR)

    # -------------------------------------------------------------------------
    def tearDown(self):
        """Undo the effects of ``pyramid.testing.setUp()``."""
        tearDown()
        if exists(TEST_DIR):
            rmtree(TEST_DIR)


# =============================================================================
class DBUnitTestCase(TestCase):
    """Base class for unit test with database use."""

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        from ..relaxng import RELAXNG
        from ..lib.breadcrumbs import Breadcrumbs
        from ..models import DB_METADATA, get_dbengine
        from ..models import get_dbsession_factory, get_tm_dbsession
        from ..models.dbsettings import DBSettings

        self.configurator = setUp(settings={
            'sqlalchemy.url': 'sqlite:///:memory:'})
        self.configurator.include('..models')

        DB_METADATA.bind = get_dbengine(self.configurator.get_settings())
        DB_METADATA.create_all()
        dbsession_factory = get_dbsession_factory(DB_METADATA.bind)
        self.configurator.registry['dbsession_factory'] = dbsession_factory
        self.configurator.registry['relaxng'] = RELAXNG
        self.configurator.registry['principals'] = []
        self.configurator.registry['themes'] = {'': {'name': {}}}
        self.configurator.registry['settings'] = \
            DBSettings.settings_defaults.copy()
        self.configurator.registry['settings']['title'] = 'TestChrysalio'

        self.request = DummyRequest(matched_route=None, is_xhr=False)
        self.request.dbsession = get_tm_dbsession(dbsession_factory, manager)
        self.request.registry = self.configurator.registry
        self.request.breadcrumbs = Breadcrumbs(self.request)
        self.request.referrer = None

    # -------------------------------------------------------------------------
    def tearDown(self):
        """Undo the effects of ``pyramid.testing.setUp()``."""
        from ..models import DB_METADATA

        tearDown()
        manager.abort()
        DB_METADATA.drop_all()
        if exists(TEST_DIR):
            rmtree(TEST_DIR)

    # -------------------------------------------------------------------------
    def add_user(self, record):
        """Add a user in the database.

        :param dict record:
            Dictionary representing the user.
        :rtype: ``None`` or :class:`pyramid.i18n.TranslationString`
        :return:
            ``None`` or error message.
        """
        error = DBUser.record_format(record)
        if error is not None:
            return error

        dbuser = DBUser(**record)
        if 'password' in record:
            dbuser.set_password(record['password'])
        # pylint: disable = no-member
        self.request.dbsession.add(dbuser)
        return None


# =============================================================================
class FunctionalTestCase(TestCase):
    """Base class for functional test."""

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        from ..relaxng import RELAXNG
        from ..scripts.ciopopulate import Populate
        from ..models.populate import xml2db

        register_loader_type(AssertionRewritingHook, DefaultProvider)

        self.testapp = BaseTestApp(
            'config:{0}'.format(TEST_INI), relative_to='.')

        args = namedtuple(
            'ArgumentParser', 'conf_uri options lang attachments')(
                conf_uri=TEST_INI, options=None, lang=None, attachments=None)
        dbsession_factory = self.testapp.app.registry['dbsession_factory']
        Populate(
            args, DBUser, xml2db, RELAXNG,
            dbsession_factory=dbsession_factory).run()

    # -------------------------------------------------------------------------
    def tearDown(self):
        """Undo the effects of ``setUp()``."""
        from ..models import DB_METADATA

        manager.abort()
        DB_METADATA.drop_all()
        if exists(TEST_DIR):
            rmtree(TEST_DIR)

    # -------------------------------------------------------------------------
    def login(self, login_, password=None):
        """Simulate a login action."""
        self.testapp.post(
            '/login',
            {'csrf_token': self.testapp.get('/login').form.get(
                'csrf_token').value,
             'login': login_,
             'password': password or '{0}pwd'.format(login_)},
            status=302)


# =============================================================================
class DummyPOST(dict):
    """Class to simulate POST with getall() method."""

    multikeys = {}

    # -------------------------------------------------------------------------
    def getall(self, key):
        """Simulation of a dict-like collection of key-value pairs where key
        might be occurred more than once in the container.

        :param str key:
            Key to retrieve.
        """
        return self.multikeys.get(key) or []


# =============================================================================
class DummyRootFactory(object):
    """Dummy Access Control List (ACL) definition."""
    # pylint: disable = too-few-public-methods
    __acl__ = [(Allow, Authenticated, 'authenticated')]

    # -------------------------------------------------------------------------
    def __init__(self, request):
        """Constructor method."""



# -*- coding: utf-8 -*-
# pylint: disable = import-outside-toplevel
"""Tests of ``views.group`` class."""

from json import dumps
from collections import namedtuple
from cgi import FieldStorage

from pyramid.response import Response
from pyramid.httpexceptions import HTTPNotFound, HTTPFound
from pyramid.testing import DummySession

from . import DBUnitTestCase, FunctionalTestCase


# =============================================================================
class UViewsGroupGroupView(DBUnitTestCase):
    """Unit test class for testing :class:`views.group.GroupView`."""

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        from ..models.dbprofile import DBProfile
        from ..models.dbuser import DBUser
        from ..models.dbgroup import DBGroup, DBGroupUser
        from . import TEST_DIR

        super(UViewsGroupGroupView, self).setUp()

        dbsession = self.request.dbsession
        self.request.registry.settings['site.uid'] = 'test'
        self.request.registry.settings['attachments'] = TEST_DIR

        dbsession.add(DBProfile(
            profile_id='profile_manager',
            i18n_label=dumps({'en': 'Profile manager'})))
        self.add_user({
            'login': 'member1', 'first_name': 'Cécile',
            'last_name': 'OURKESSA', 'password': 'member1pwd',
            'email': 'member1@chrysal.io'})
        self.add_user({
            'login': 'test1', 'first_name': 'Édith', 'last_name': 'AVULEUR',
            'password': 'test1pwd', 'email': 'test1@chrysal.io'})
        user_id = dbsession.query(DBUser.user_id).filter_by(
            login='member1').first()[0]
        dbgroup = DBGroup(
            group_id='managers',
            i18n_label=dumps({'en': 'Managers'}),
            i18n_description={'en': 'Permission to view, edit and create.'},
            attachments_key='Managers', picture='managers.png')
        dbgroup.users.append(DBGroupUser(user_id=user_id))
        dbgroup.profiles.append(DBProfile(
            profile_id='user_manager',
            i18n_label=dumps({'en': 'User account manager'})))
        dbsession.add(dbgroup)
        dbsession.add(DBGroup(
            group_id='editors', i18n_label=dumps({'en': 'Editors'})))
        self.request.session['user'] = {
            'login': 'member1', 'user_id': user_id}

        self.configurator.add_route('home', '/')
        self.configurator.add_route('group_index', '/group/index')
        self.configurator.add_route('group_create', '/group/create')
        self.configurator.add_route('group_view', '/group/view/{group_id}')
        self.configurator.add_route('group_edit', '/group/edit/{group_id}')
        self.configurator.add_route('attachment', '/attachment/*path')

    # -------------------------------------------------------------------------
    def test_index(self):
        """[u:views.group.GroupView.index]"""
        from ..lib.filter import Filter
        from ..views.group import GroupView
        from . import DummyPOST

        self.request.matched_route = namedtuple('Route', 'name')(
            name='group_index')
        self.request.session = DummySession({'user': {'user_id': 1}})

        # As creator
        response = GroupView(self.request).index()
        self.assertIn('pfilter', response)
        self.assertIsInstance(response['pfilter'], Filter)
        self.assertTrue(response['pfilter'].is_empty())
        self.assertIn('paging', response)
        self.assertIn('PAGE_SIZES', response)
        self.assertIn('i_creator', response)
        self.assertTrue(response['i_creator'])
        self.assertTrue(response['i_editor'])
        self.assertEqual(self.request.breadcrumbs.current_title(), 'Groups')

        # Not as creator
        self.request.has_permission = lambda k: False
        response = GroupView(self.request).index()
        self.assertFalse(response['i_creator'])
        self.assertFalse(response['i_editor'])

        # AJAX mode
        self.request.has_permission = lambda k: True
        self.request.is_xhr = True
        self.request.POST = DummyPOST()
        response = GroupView(self.request).index()
        self.assertIsInstance(response, dict)
        self.assertFalse(response)

    # -------------------------------------------------------------------------
    def test_index_delete(self):
        """[u:views.group.GroupView.index] delete group"""
        from ..models.dbgroup import DBGroup
        from ..views.group import GroupView

        dbsession = self.request.dbsession
        self.assertEqual(dbsession.query(DBGroup).count(), 2)

        self.request.POST = {'del!editors.x': True}
        GroupView(self.request).index()
        self.assertEqual(dbsession.query(DBGroup).count(), 1)

    # -------------------------------------------------------------------------
    def test_index_import(self):
        """[u:views.group.GroupView.index] import group"""
        from . import TEST1_GROUP_XML, DummyPOST
        from ..views.group import GroupView

        self.request.POST = DummyPOST({'imp!.x': True})
        response = GroupView(self.request).index()
        self.assertIn('paging', response)
        self.assertNotIn('team1', [k.group_id for k in response['paging']])

        with open(TEST1_GROUP_XML, 'r', encoding='utf8') as hdl:
            input_file = FieldStorage()
            input_file.file = hdl
            input_file.filename = TEST1_GROUP_XML
            self.request.POST.multikeys = {'file': (input_file,)}
            response = GroupView(self.request).index()
        self.assertIn('paging', response)
        self.assertIn('team1', [k.group_id for k in response['paging']])

    # -------------------------------------------------------------------------
    def test_index_export(self):
        """[u:views.group.GroupView.index] export group"""
        from ..views.group import GroupView

        self.request.POST = {'exp!foo.x': True}
        response = GroupView(self.request).index()
        self.assertIsInstance(response, dict)

        self.request.POST = {'exp!managers.x': True}
        response = GroupView(self.request).index()
        self.assertIsInstance(response, Response)
        self.assertEqual(response.content_type, 'application/xml')
        self.assertIn(b'managers', response.body)

    # -------------------------------------------------------------------------
    def test_index_filter(self):
        """[u:views.group.GroupView.index_filter]"""
        from ..views.group import GroupView

        self.request.is_xhr = True
        self.request.params = {'field': 'i18n_label', 'value': 'Managers'}
        data = GroupView(self.request).index_filter()
        self.assertIsInstance(data, list)
        self.assertIn('Managers', data)

    # -------------------------------------------------------------------------
    def test_view(self):
        """[u:views.group.GroupView.view]"""
        from ..views.group import GroupView

        self.request.matched_route = namedtuple('Route', 'name')(
            name='group_view')

        self.request.matchdict = {'group_id': 'foo'}
        self.assertRaises(HTTPNotFound, GroupView(self.request).view)

        self.request.matchdict = {'group_id': 'managers'}
        response = GroupView(self.request).view()
        self.assertIn('form', response)
        self.assertEqual(
            self.request.breadcrumbs.current_title(), 'Group "Managers"')

    # -------------------------------------------------------------------------
    def test_view_export(self):
        """[u:views.group.GroupView.view] export groups"""
        from ..views.group import GroupView

        self.request.POST = {'exp!.x': True}
        self.request.matchdict = {'group_id': 'managers'}

        response = GroupView(self.request).view()
        self.assertIsInstance(response, Response)
        self.assertEqual(response.content_type, 'application/xml')
        self.assertIn(b'managers', response.body)

    # -------------------------------------------------------------------------
    def test_create(self):
        """[u:views.group.GroupView.edit] creation"""
        from ..views.group import GroupView

        self.request.matched_route = namedtuple('Route', 'name')(
            name='group_create')

        response = GroupView(self.request).edit()
        self.assertIn('form', response)
        self.assertIn('dbgroup', response)
        self.assertIn('user_filter', response)
        self.assertIn('user_paging', response)
        self.assertIn('picture', response)
        self.assertIn('label', response)
        self.assertIsNone(response['label'])
        self.assertEqual(
            self.request.breadcrumbs.current_title(), 'Group Creation')

    # -------------------------------------------------------------------------
    def test_create_save(self):
        """[u:views.group.GroupView.edit] save creation"""
        from ..views.group import GroupView

        self.request.matched_route = namedtuple('Route', 'name')(
            name='group_create')
        self.request.POST = {
            'group_id': 'managers', 'label_en': 'Managers', 'sav!.x': True}

        # Existing group
        response = GroupView(self.request).edit()
        errors = self.request.session.pop_flash('alert')
        self.assertEqual(len(errors), 1)
        self.assertIn('form', response)
        self.assertIn('exists', errors[0])
        self.assertEqual(
            self.request.breadcrumbs.current_title(), 'Group Creation')
        # pylint: disable = no-member
        self.request.dbsession.rollback()
        # pylint: enable = no-member

        # New group, missing label
        self.request.POST = {'group_id': 'team1', 'sav!.x': True}
        GroupView(self.request).edit()
        errors = self.request.session.pop_flash('alert')
        self.assertEqual(len(errors), 1)
        self.assertIn('Correct', errors[0])
        # pylint: disable = no-member
        self.request.dbsession.rollback()
        # pylint: enable = no-member

        # New group OK
        self.add_user({
            'login': 'member1', 'first_name': 'Cécile',
            'last_name': 'OURKESSA', 'password': 'member1pwd',
            'email': 'member1@chrysal.io'})
        self.request.POST['label_en'] = 'Team 1'
        GroupView(self.request).edit()
        self.assertFalse(self.request.session.pop_flash('alert'))

    # -------------------------------------------------------------------------
    def test_edit(self):
        """[u:views.group.GroupView.edit]"""
        from ..models.dbgroup import DBGroup
        from ..views.group import GroupView

        self.request.matched_route = namedtuple('Route', 'name')(
            name='group_edit')
        self.request.matchdict = {'group_id': 'managers'}

        response = GroupView(self.request).edit()
        self.assertIn('form', response)
        self.assertIn('label', response)
        self.assertEqual(response['label'], 'Managers')
        self.assertIn('dbgroup', response)
        self.assertIsInstance(response['dbgroup'], DBGroup)
        self.assertEqual(
            self.request.breadcrumbs.current_title(),
            'Group "Managers" Edition')

    # -------------------------------------------------------------------------
    def test_edit_save(self):
        """[u:views.group.GroupView.edit] save"""
        from ..models.dbuser import DBUser
        from ..views.group import GroupView

        dbsession = self.request.dbsession
        self.request.matchdict = {'group_id': 'managers'}
        self.request.POST = {
            'description_en': 'Managers can create, destroy and modify.',
            'sav!.x': True}

        # Missing label
        GroupView(self.request).edit()
        errors = self.request.session.pop_flash('alert')
        self.assertEqual(len(errors), 1)
        self.assertIn('Correct', errors[0])

        # Correct group
        user1_id = dbsession.query(DBUser.user_id).filter_by(
            login='member1').first()[0]
        user2_id = dbsession.query(DBUser.user_id).filter_by(
            login='test1').first()[0]
        self.request.POST['label_en'] = 'Managers'
        self.request.POST['shw:{0}'.format(user1_id)] = True
        self.request.POST['shw:{0}'.format(user2_id)] = True
        self.request.POST['mbr:{0}'.format(user2_id)] = True
        self.request.POST['pfl:profile_manager'] = True
        self.assertIsInstance(GroupView(self.request).edit(), HTTPFound)

        # Whitout permission
        self.request.has_permission = lambda k: False
        self.assertIsInstance(GroupView(self.request).edit(), HTTPFound)

    # -------------------------------------------------------------------------
    def test_edit_picture(self):
        """[u:views.group.GroupView.edit] change picture"""
        from ..views.group import GroupView
        from . import TEST1_SVG

        self.request.matchdict = {'group_id': 'managers'}
        self.request.POST = {'pct!.x': True}

        # Normal HTTP
        with open(TEST1_SVG, 'rb') as hdl:
            input_file = FieldStorage()
            input_file.filename = TEST1_SVG
            input_file.file = hdl
            self.request.POST['picture'] = input_file
            GroupView(self.request).edit()

        # AJAX
        with open(TEST1_SVG, 'rb') as hdl:
            input_file = FieldStorage()
            input_file.filename = TEST1_SVG
            input_file.file = hdl
            self.request.is_xhr = True
            self.request.POST['picture'] = input_file
            response = GroupView(self.request).edit()
        self.assertIsInstance(response, dict)
        self.assertFalse(response)


# =============================================================================
class FViewsGroupGroupView(FunctionalTestCase):
    """Functional test class for testing :class:`views.group.GroupView`."""

    # -------------------------------------------------------------------------
    def test_index(self):
        """[f:views.group.GroupView.index]"""
        self.login('test1')
        self.testapp.get('/group/index', status=403)

        self.login('admin')
        response = self.testapp.get('/group/index', status=200)
        self.assertIn(b'Groups', response.body)
        self.assertIn(b'Create', response.body)
        self.assertIn(b'Import', response.body)
        self.assertIn(b'Export selected group', response.body)
        self.assertIn(b'Delete selected group', response.body)

    # -------------------------------------------------------------------------
    def test_view(self):
        """[f:views.group.GroupView.view]"""
        self.login('test1')
        self.testapp.get('/group/view/team1', status=403)

        self.login('admin')
        response = self.testapp.get('/group/view/team1', status=200)
        self.assertIn(b'Edit', response.body)

    # -------------------------------------------------------------------------
    def test_create(self):
        """[f:views.group.GroupView.create]"""
        self.login('test1')
        self.testapp.get('/group/create', status=403)

        self.login('admin')
        response = self.testapp.get('/group/create', status=200)
        self.assertIn(b'Create', response.body)

    # -------------------------------------------------------------------------
    def test_edit(self):
        """[f:views.group.GroupView.edit]"""
        self.login('test1')
        self.testapp.get('/group/edit/team1', status=403)

        self.login('admin')
        response = self.testapp.get('/group/edit/team1', status=200)
        self.assertIn(b'Save', response.body)



"""Functions to manage RESTful API."""

from datetime import datetime
from json import loads, dumps
from math import fabs
from http.client import BadStatusLine

from future.moves.urllib.parse import urlencode
from future.moves.urllib.request import urlopen
from future.moves.urllib.error import URLError

from .i18n import _
from .utils import encrypt, decrypt
from ..models.dbuser import DBUser


TOKEN_TTL = 1.0


# =============================================================================
def restful_call(url, login, key, data=None):
    """RESTful call.

    :param str url:
        RESTful URL to call.
    :param str login:
        Login of an authorized user.
    :param str key:
        Key to decrypt the token.
    :param dict data: (optional)
        Data to transfer.
    :rtype: tuple
    :return:
        A tuple such as ``(response, error)``.
    """
    if data is None:
        data = {}
    for k in data:
        if isinstance(data[k], (tuple, list, dict)):
            data[k] = dumps(data[k])
    data['login'] = login
    data['token'] = encrypt(datetime.utcnow().isoformat(), key or '-')
    data = urlencode(data).encode('ascii')

    try:
        with urlopen(url, data) as hdl:
            response = loads(hdl.read().decode('utf8'))
    except (URLError, ValueError, BadStatusLine) as error:
        return None, str(error)

    if response['error'] is not None:
        return None, response['error']

    return response, None


# =============================================================================
def restful_login(request, key, token_ttl=None):
    """Login during a RESTful call.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    :param str key:
        Key to decrypt the token.
    :param float token_ttl: (optional)
        Validity period in seconds for the token.
    :rtype: str
    :return:
        Translated error message or ``None``.
    """
    if 'user' in request.session:
        return None

    # Check the token
    token = request.params.get('token')
    if not token:
        return request.localizer.translate(_('Token is missing.'))
    token = decrypt(token, key or '-')
    if not token:
        return request.localizer.translate(_('Token is invalid.'))
    token = datetime.strptime(token, '%Y-%m-%dT%H:%M:%S.%f')
    delta = (datetime.utcnow() - token).total_seconds()
    if fabs(delta) > float(token_ttl or TOKEN_TTL):
        return request.localizer.translate(_('Token has expired.'))

    # Set up user session
    dbuser = DBUser.get(request, request.params.get('login'))[0]
    if dbuser is None:
        return request.localizer.translate(_('Access denied.'))
    dbuser.set_session(request)

    return None



"""Attachment view callables."""

from os import sep
from os.path import join, dirname

from pyramid.view import view_config
from pyramid.httpexceptions import HTTPNotFound
from pyramid.response import FileResponse
from pyramid.asset import abspath_from_asset_spec
from pyramid.security import NO_PERMISSION_REQUIRED


# =============================================================================
@view_config(route_name='attachment')
def attachment(request):
    """Output an attachment file.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    """
    attachments = request.registry.settings.get('attachments')
    if not attachments:
        raise HTTPNotFound()

    try:
        return FileResponse(
            join(attachments, sep.join(request.matchdict['path'])),
            request=request, cache_max_age=3600)
    except OSError:
        raise HTTPNotFound()


# =============================================================================
@view_config(route_name='favicon', permission=NO_PERMISSION_REQUIRED)
def favicon(request):
    """Output the `favicon.ico` file.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    """
    settings = request.registry.settings
    if settings.get('site.favicon'):
        icon = abspath_from_asset_spec(settings['site.favicon'])
    else:
        icon = join(dirname(__file__), '..', 'Static', 'favicon.ico')
    return FileResponse(icon, request=request, cache_max_age=3600)


# =============================================================================
@view_config(route_name='robots', permission=NO_PERMISSION_REQUIRED)
def robots(request):
    """Output the `robots.txt` file.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    """
    settings = request.registry.settings
    if settings.get('site.robots'):
        robots_file = abspath_from_asset_spec(settings['site.robots'])
    else:
        robots_file = join(dirname(__file__), '..', 'Static', 'robots.txt')
    return FileResponse(robots_file, request=request, cache_max_age=3600)



# -*- coding: utf-8 -*-
# pylint: disable = too-many-lines
"""SQLAlchemy-powered model definitions for users."""

from datetime import datetime, date
from bcrypt import hashpw, gensalt

import colander
from sqlalchemy import Column, Integer, ForeignKey
from sqlalchemy import String, Enum, Date, DateTime, Boolean
from sqlalchemy.orm import relationship
from lxml import etree
from webhelpers2.html import HTML, literal
from pytz import common_timezones

from ..lib.i18n import _
from ..lib.config import settings_get_list
from ..lib.utils import make_id, normalize_spaces, age
from ..lib.mailing import Mailing
from ..lib.form import SameAs
from ..lib.paging import PAGE_SIZES
from ..lib.attachment import attachment_url
from . import DBDeclarativeClass, ID_LEN, NAME_LEN, LABEL_LEN, EMAIL_LEN
from .dbbase import DBBaseClass
from .dbprofile import DBProfile, DBProfilePrincipal

USER_STATUS_LABELS = {
    'administrator': _('administrator'), 'active': _('active'),
    'locked': _('locked'), 'inactive': _('inactive')}
USER_HONORIFIC_LABELS = {'Mr': _('Mr'), 'Mrs': _('Mrs')}


# =============================================================================
class DBUser(DBDeclarativeClass, DBBaseClass):
    """SQLAlchemy-powered user model."""

    status_labels = USER_STATUS_LABELS
    honorific_labels = USER_HONORIFIC_LABELS
    suffix = 'ciousr'
    attachments_dir = 'Users'
    _settings_tabs = (
        _('Information'), _('Preferences'), _('Profiles'), _('Groups'))

    __tablename__ = 'users'
    __table_args__ = {'mysql_engine': 'InnoDB'}

    user_id = Column(Integer, autoincrement=True, primary_key=True)
    login = Column(String(EMAIL_LEN), unique=True, index=True, nullable=False)
    status = Column(
        Enum(*USER_STATUS_LABELS.keys(), name='status'), default='active')
    password = Column(String(64), nullable=False)
    password_update = Column(DateTime)
    password_mustchange = Column(
        Boolean(name='password_mustchange'), default=False)
    first_name = Column(String(NAME_LEN))
    last_name = Column(String(NAME_LEN), nullable=False)
    honorific = Column(
        Enum(*USER_HONORIFIC_LABELS.keys(), name='honorific'))
    email = Column(String(EMAIL_LEN), nullable=False)
    email_hidden = Column(Boolean(name='email_hidden'), default=False)
    language = Column(String(5))
    time_zone = Column(String(48))
    theme = Column(String(ID_LEN))
    page_size = Column(Integer)
    attachments_key = Column(String(ID_LEN + 20))
    picture = Column(String(ID_LEN + 4))
    expiration = Column(Date)
    last_login = Column(DateTime)
    account_update = Column(DateTime)
    account_creation = Column(DateTime, default=datetime.now)
    authority = Column(String(ID_LEN))
    authority_check = Column(DateTime)

    profiles = relationship(DBProfile, secondary='users_profiles')
    groups = relationship('DBGroup', secondary='groups_users', viewonly=True)

    # -------------------------------------------------------------------------
    def set_password(self, password):
        """Set the password, possibly hashing it.

        :param str password:
            Password to set. If it does not begin with ``$``, we use bcrypt
            algorithm before setting.
        """
        if not password.startswith('$'):
            self.password = hashpw(
                password.encode('utf8'), gensalt()).decode('utf8')
        else:
            self.password = password
        self.password_update = datetime.now()

    # -------------------------------------------------------------------------
    def check_password(self, password):
        """Check the validy of the given password.

        :param str password:
            Clear password to check.
        :rtype: bool
        """
        if password and self.password is not None:
            expected = self.password.encode('utf8')
            return expected == hashpw(password.encode('utf8'), expected)
        return False

    # -------------------------------------------------------------------------
    @classmethod
    def get(cls, request, login=None, password=None):
        """Retrieve a user using ``login`` and ``password`` or the content of
        ``request.params``.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :param str login: (optional)
            Login of the user to authenticate. If it is ``None``, we try to
            find it in ``request.params``.
        :param str password: (optional)
            Clear password.  If it is ``None``, we try to find it in
            ``request.params``.
        :rtype: tuple
        :return:
            A tuple like ``(dbuser, error)`` where ``dbuser`` is a DBUser
            object representing the authenticated user or ``None`` and
            ``error`` is an error message.

        If ``password`` is ``None`` and not in ``request.params``, password
        checking is not performed.

        If the user is authenticated, it updates ``last_login`` field in
        database.
        """
        # pylint: disable = too-many-return-statements
        login = login or request.params.get('login')
        if not login:
            return None, _('ID or password is incorrect.')
        password = password or request.params.get('password')
        dbuser = request.dbsession.query(cls).filter_by(
            login=make_id(login, 'token', EMAIL_LEN)).first()

        if dbuser is None:
            if 'authorities' in request.registry:
                for authority in request.registry['authorities'].values():
                    dbuser, error = authority.get(
                        request, login, password, cls)
                    if dbuser is not None or error is not None:
                        return dbuser, error
            return None, _('ID or password is incorrect.')

        if dbuser.authority:
            if 'authorities' not in request.registry or \
               dbuser.authority not in request.registry['authorities']:
                return None, _('Authority not available.')
            error = request.registry['authorities'][
                dbuser.authority].check(request, login, password, dbuser)
            if error:
                return None, error
            password = None

        if password is not None and not dbuser.check_password(password):
            return None, _('ID or password is incorrect.')
        if dbuser.status == 'locked':
            return None, _('Your account is locked.')
        if dbuser.status not in ('administrator', 'active'):
            return None, _('Your account is not active.')
        if dbuser.status != 'administrator' and dbuser.expiration \
           and dbuser.expiration < date.today():
            return None, _('Your account has expired.')

        dbuser.last_login = datetime.now()
        return dbuser, None

    # -------------------------------------------------------------------------
    def set_session(self, request):
        """Set up user session (``session['user']``).

        :type  request: pyramid.request.Request
        :param request:
            Current request.

        It saves in session the following values:

        * ``lang``: user language
        * ``theme``: user theme
        * ``user``: user dictionary

        The user dictionary includes:

        * ``user_id``: user ID
        * ``login``: user login
        * ``email``: user e-mail
        * ``name``: user first name and last name
        * ``attachment``: attachment ID
        * ``principals``: list of principals (premission groups) (see
          ":ref:`Pyramid Security <pyramid:security_chapter>`" for more
          information)
        """
        # Language
        settings = request.registry.settings
        if self.language:
            langs = settings_get_list(settings, 'languages', ('en',))
            request.session['lang'] = \
                (self.language in langs and self.language) or \
                (self.language[0:2] in langs and self.language[0:2]) or \
                request.registry['settings']['language']

        # Theme
        if self.theme and self.theme in request.registry['themes']:
            request.session['theme'] = self.theme
        else:
            request.session['theme'] = request.registry['settings']['theme']

        # Groups
        groups = set()
        principals = set()
        for dbgroup in self.groups:
            groups.add(dbgroup.group_id)
            for dbprofile in dbgroup.profiles:
                principals |= {k.principal for k in dbprofile.principals}
        groups = tuple(groups)

        # Principals
        if self.status == 'administrator':
            principals = ('system.administrator',)
        else:
            principals |= {
                k.principal for k in
                request.dbsession.query(DBProfilePrincipal)
                .join((DBProfile, DBUser.profiles))
                .join((DBProfilePrincipal, DBProfile.principals))
                .filter(DBUser.user_id == self.user_id)}
            principals = tuple(principals)

        # User information
        request.session['user'] = {
            'user_id': self.user_id,
            'login': self.login,
            'email': self.email,
            'name': '{0} {1}'.format(
                self.first_name or '', self.last_name).strip(),
            'picture': attachment_url(
                request, 'Users', self.attachments_key, self.picture),
            'principals': principals,
            'groups': groups}

        # Reset paging
        request.session['paging'] = (
            self.page_size if self.page_size else
            request.registry['settings']['page-size'], {})

        # Reset modes and menu
        if 'modes' in request.session:
            request.session['modes'][2] = None
        if 'menu' in request.session:
            del request.session['menu']

    # -------------------------------------------------------------------------
    @classmethod
    def load_administrator(cls, dbsession, record):
        """Load the administrator user from INI configuration file.

        :type  dbsession: sqlalchemy.orm.session.Session
        :param dbsession:
            SQLAlchemy session.
        :param dict record:
            Dictionary representing the administrator configuration.
        :rtype: ``None`` or :class:`pyramid.i18n.TranslationString`
        :return:
            ``None`` or error message.
        """
        # Check user existence
        if not record.get('login'):
            return _('Login is missing.')
        login = make_id(record['login'], 'token', EMAIL_LEN)
        dbuser = dbsession.query(cls.login).filter_by(login=login).first()
        if dbuser is not None:
            return None

        # Complete record
        record['status'] = 'administrator'
        record['email_hidden'] = True

        # Check final record and create user
        error = cls.record_format(record)
        if error:
            return error
        dbuser = cls(**record)

        dbsession.add(dbuser)
        return None

    # -------------------------------------------------------------------------
    @classmethod
    def xml2db(cls, dbsession, user_elt, error_if_exists=True, kwargs=None):
        """Load a user from a XML element.

        :type  dbsession: sqlalchemy.orm.session.Session
        :param dbsession:
            SQLAlchemy session.
        :type  user_elt: lxml.etree.Element
        :param user_elt:
            User XML element.
        :param bool error_if_exists: (default=True)
            It returns an error if user already exists.
        :param dict kwargs: (optional)
            Dictionary of keyword arguments with the key ``'profiles'``.
        :rtype: :class:`pyramid.i18n.TranslationString` or ``None``
        :return:
            Error message or ``None``.
        """
        # Check if already exists
        login = make_id(user_elt.findtext('login'), 'token', EMAIL_LEN)
        dbuser = dbsession.query(cls).filter_by(login=login).first()
        if dbuser is not None:
            if error_if_exists:
                return _('User "${l}" already exists.', {'l': login})
            return None

        # Create user
        record = cls.record_from_xml(login, user_elt)
        error = cls.record_format(record)
        if error:
            return error
        dbuser = cls(**record)
        dbsession.add(dbuser)

        # Fill extra tables
        return dbuser.xml2db_extra(dbsession, user_elt, kwargs)

    # -------------------------------------------------------------------------
    def xml2db_extra(self, dbsession, user_elt, kwargs):
        """Load extra information on a user from a XML element.

        :type  dbsession: sqlalchemy.orm.session.Session
        :param dbsession:
            SQLAlchemy session.
        :type  user_elt: lxml.etree.Element
        :param user_elt:
            User XML element.
        :param dict kawrgs:
            Dictionary of keyword arguments with the key ``'profiles'``.
        :rtype: :class:`pyramid.i18n.TranslationString` or ``None``
        :return:
            Error message or ``None``.
        """
        dbsession.flush()

        # Add profiles
        profile_ids = kwargs.get('profiles') if kwargs else None
        if profile_ids:
            done = set()
            for elt in user_elt.findall('profiles/profile'):
                profile_id = elt.text
                if profile_id not in done and profile_id in profile_ids:
                    dbsession.add(DBUserProfile(
                        user_id=self.user_id, profile_id=profile_id))
                    done.add(profile_id)

    # -------------------------------------------------------------------------
    @classmethod
    def record_from_xml(cls, login, user_elt):
        """Convert an user XML element into a dictionary.

        :param str login:
            User login.
        :type  user_elt: lxml.etree.Element
        :param user_elt:
            User XML element.
        :rtype: dict
        """
        password_elt = user_elt.find('password')
        email_elt = user_elt.find('email')
        attachments_elt = user_elt.find('attachments')
        record = {
            'login': login,
            'status': user_elt.get('status', 'active'),
            'password': password_elt is not None and password_elt.text,
            'password_update':
            password_elt is not None and password_elt.get('updated'),
            'password_mustchange':
            password_elt is not None and
            password_elt.get('must-change') == 'true' or False,
            'first_name': user_elt.findtext('firstname'),
            'last_name': user_elt.findtext('lastname'),
            'honorific': user_elt.findtext('honorific'),
            'email': email_elt is not None and email_elt.text,
            'email_hidden': email_elt is not None and email_elt.get(
                'hidden') == 'true' or False,
            'language': user_elt.findtext('language'),
            'time_zone': user_elt.findtext('timezone'),
            'theme': user_elt.findtext('theme'),
            'attachments_key':
            attachments_elt is not None and attachments_elt.get('key') or None,
            'picture':
            attachments_elt is not None and attachments_elt.findtext(
                'picture') or None,
            'page_size': user_elt.findtext('page-size'),
            'expiration': user_elt.findtext('expiration'),
            'last_login': user_elt.get('last-login'),
            'account_update': user_elt.get('updated'),
            'account_creation': user_elt.get('created'),
            'authority': user_elt.findtext('authority'),
            'authority_check': user_elt.find('authority').get(
                'checked') if user_elt.find('authority') is not None else None}
        return record

    # -------------------------------------------------------------------------
    @classmethod
    def record_format(cls, record):
        """Check and possibly correct a record before inserting it in the
        database.

        :param dict record:
            Dictionary of values to check.
        :rtype: ``None`` or :class:`pyramid.i18n.TranslationString`
        :return:
            ``None`` or error message.
        """
        # pylint: disable = too-many-return-statements
        for k in [i for i in record if record[i] is None]:
            del record[k]

        # Login
        if not record.get('login'):
            return _('User without login.')
        record['login'] = make_id(record['login'], 'token', EMAIL_LEN)

        # Password
        if not record.get('password'):
            return _('User "${l}" without password.', {'l': record['login']})
        if not record['password'].startswith('$'):
            record['password'] = hashpw(
                record['password'].encode('utf8'), gensalt()).decode('utf8')

        # Name
        if not record.get('last_name'):
            return _('User "${l}" without last name.', {'l': record['login']})
        record['last_name'] = normalize_spaces(record['last_name'], NAME_LEN)
        if record.get('first_name'):
            record['first_name'] = normalize_spaces(
                record['first_name'], NAME_LEN)
        if (record.get('honorific') or 'Mr') \
           not in cls.honorific.property.columns[0].type.enums:
            return _('User "${l}" with unknown honorific title.',
                     {'l': record['login']})

        # Email
        if not record.get('email'):
            return _('User "${l}" without email.', {'l': record['login']})
        record['email'] = record['email'][0:EMAIL_LEN]
        if not Mailing.is_valid(record['email']):
            return _('Invalid email address: ${e}', {'e': record['email']})

        # Language
        if record.get('language'):
            record['language'] = record['language'][0:5]

        # Authority
        if record.get('authority_check') and not record.get('authority'):
            del record['authority_check']

        # Dates
        cls.record_convert_dates(record)
        return None

    # -------------------------------------------------------------------------
    @classmethod
    def record_convert_dates(cls, record):
        """Possibly convert dates of a record.

        :param dict record:
            Dictionary of values to check.
        """
        if 'last_login' in record and \
           not isinstance(record['last_login'], datetime):
            record['last_login'] = datetime.strptime(
                record['last_login'], '%Y-%m-%dT%H:%M:%S')

        if 'password_update' in record and \
           not isinstance(record['password_update'], datetime):
            record['password_update'] = datetime.strptime(
                record['password_update'], '%Y-%m-%dT%H:%M:%S')

        if 'expiration' in record and \
           not isinstance(record['expiration'], (date, datetime)):
            record['expiration'] = datetime.strptime(
                record['expiration'], '%Y-%m-%d').date()

        if 'account_creation' in record and \
           not isinstance(record['account_creation'], datetime):
            record['account_creation'] = datetime.strptime(
                record['account_creation'], '%Y-%m-%dT%H:%M:%S')

        if 'account_update' in record and \
           not isinstance(record['account_update'], datetime):
            record['account_update'] = datetime.strptime(
                record['account_update'], '%Y-%m-%dT%H:%M:%S')

        if 'authority_check' in record and \
           not isinstance(record['authority_check'], datetime):
            record['authority_check'] = datetime.strptime(
                record['authority_check'], '%Y-%m-%dT%H:%M:%S')

    # -------------------------------------------------------------------------
    def db2xml(self, dbsession=None):  # noqa
        """Serialize a user to a XML representation.

        :type  dbsession: sqlalchemy.orm.session.Session
        :param dbsession: (optional)
            SQLAlchemy session.
        :rtype: lxml.etree.Element
        """
        # pylint: disable = too-many-branches, unused-argument
        user_elt = etree.Element('user')
        if self.status != 'active':
            user_elt.set('status', self.status)
        user_elt.set(
            'created', self.account_creation.isoformat().partition('.')[0])
        if self.account_update and \
           self.account_update != self.account_creation:
            user_elt.set(
                'updated', self.account_update.isoformat().partition('.')[0])
        if self.last_login:
            user_elt.set(
                'last-login', self.last_login.isoformat().partition('.')[0])

        etree.SubElement(user_elt, 'login').text = self.login

        elt = etree.SubElement(user_elt, 'password')
        elt.text = self.password
        if self.password_update:
            elt.set(
                'updated', self.password_update.isoformat().partition('.')[0])
        if self.password_mustchange:
            elt.set('must-change', 'true')

        if self.first_name:
            etree.SubElement(user_elt, 'firstname').text = self.first_name
        etree.SubElement(user_elt, 'lastname').text = self.last_name
        if self.honorific:
            etree.SubElement(user_elt, 'honorific').text = self.honorific

        elt = etree.SubElement(user_elt, 'email')
        elt.text = self.email
        if self.email_hidden:
            elt.set('hidden', 'true')

        if self.language:
            etree.SubElement(user_elt, 'language').text = self.language
        if self.time_zone:
            etree.SubElement(user_elt, 'timezone').text = self.time_zone
        if self.theme:
            etree.SubElement(user_elt, 'theme').text = self.theme
        if self.attachments_key:
            elt = etree.SubElement(
                user_elt, 'attachments', key=self.attachments_key)
            if self.picture:
                etree.SubElement(elt, 'picture').text = self.picture
        if self.page_size:
            etree.SubElement(user_elt, 'page-size').text = str(self.page_size)
        if self.expiration:
            etree.SubElement(user_elt, 'expiration').text = \
                self.expiration.isoformat()

        if self.authority:
            elt = etree.SubElement(user_elt, 'authority')
            elt.text = self.authority
            if self.authority_check:
                elt.set(
                    'checked',
                    self.authority_check.isoformat().partition('.')[0])

        self.db2xml_extra(user_elt)

        return user_elt

    # -------------------------------------------------------------------------
    def db2xml_extra(self, user_elt):
        """Serialize depedencies of a user to a XML representation.

        :type  user_elt: lxml.etree.Element
        :param user_elt:
            User XML element.
        """
        # Profiles
        if self.profiles:
            elt = etree.SubElement(user_elt, 'profiles')
            for item in self.profiles:
                etree.SubElement(elt, 'profile').text = item.profile_id

    # -------------------------------------------------------------------------
    @classmethod
    def paging_filter(cls, request, form, user_filter, user_paging):
        """Filter for users.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :type  form: .lib.form.Form
        :param form:
            Current form object.
        :type  user_filter: .lib.filter.Filter
        :param user_filter:
            Filter for users.
        :type  user_paging: .lib.paging.Paging
        :param user_paging:
            Paging for users.
        :rtype: str
        """
        translate = request.localizer.translate
        html = '<div class="cioPagingPagerFilter cioFlexRowContainer">\n'

        html += '<div class="cioPagingPager cioFlexItem">'\
            '{page_size} {pager}</div>\n'.format(
                page_size=form.select(
                    'page_size', '', PAGE_SIZES, True,
                    title=translate(_('Lines per page'))),
                pager=user_paging.pager_top())

        html += '<div class="cioPagingFilter">'
        if not user_filter.is_empty():
            html += '<span class="cioFilterConditions">{0}</span>'.format(
                user_filter.html_conditions())
        html += '{inputs} <span class="cioFilterSubmit">{submit}</span>'\
            '</div>\n'.format(
                inputs=user_filter.html_inputs(form),
                submit=form.submit(
                    'filter', translate(_('Filter')),
                    class_='cioFilterButton'))

        html += '</div>\n'
        return html

    # -------------------------------------------------------------------------
    def tab4view(self, request, tab_index, form):
        """Generate the tab content of user account.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :param int index:
            Index of the tab.
        :type  form: .lib.form.Form
        :param form:
            Current form object.
        :rtype: webhelpers2.html.literal
        """
        if tab_index == 0:
            return self._tab4view_information(request, form)
        if tab_index == 1:
            return self._tab4view_preferences(request, form)
        if tab_index == 2:
            return self._tab4view_profiles(request)
        if tab_index == 3:
            return self._tab4view_groups(request)
        return ''

    # -------------------------------------------------------------------------
    def _tab4view_information(self, request, form):
        """Generate the information tab.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :type  form: .lib.form.Form
        :param form:
            Current form object.
        :rtype: webhelpers2.html.literal
        """
        # Identify
        i_editor = request.has_permission('user-edit') or \
            self.user_id == request.session['user']['user_id']
        translate = request.localizer.translate
        html = HTML.h3(translate(_('Identity')))
        html += form.grid_item(translate(_('Login:')), self.login, clear=True)
        html += form.grid_item(
            translate(_('Title:')),
            translate(USER_HONORIFIC_LABELS.get(self.honorific))
            if self.honorific else None, clear=True)
        html += form.grid_item(
            translate(_('First Name:')), self.first_name, clear=True)
        html += form.grid_item(
            translate(_('Last Name:')), self.last_name, clear=True)
        if not self.email_hidden or i_editor:
            html += form.grid_item(
                translate(_('Email:')), self.email, clear=True)
        html += form.grid_item(
            translate(_('Language:')), self.language, clear=True)
        html += form.grid_item(
            translate(_('Time Zone:')), self.time_zone, clear=True)

        # Security
        html += HTML.h3(translate(_('Security')))
        html += form.grid_item(
            translate(_('Authority:')), self.authority, clear=True)
        html += form.grid_item(
            translate(_('Status:')),
            translate(self.status_labels[self.status]), clear=True)
        html += form.grid_item(
            translate(_('Creation:')), translate(age(self.account_creation)),
            title=self.account_creation.isoformat(' ').partition('.')[0],
            clear=True)
        html += form.grid_item(
            translate(_('Password:')),
            self.password_mustchange and translate(
                _('changed during next connexion')), clear=True)
        if self.account_update:
            html += form.grid_item(
                translate(_('Update:')), translate(age(self.account_update)),
                title=self.account_update.isoformat(' ').partition('.')[0],
                clear=True)
        if self.expiration:
            html += form.grid_item(
                translate(_('Expiration:')), self.expiration.isoformat(),
                clear=True)
        if self.last_login:
            html += form.grid_item(
                translate(_('Last connection:')),
                translate(age(self.last_login)),
                title=self.last_login.isoformat(' ').partition('.')[0],
                clear=True)

        return html

    # -------------------------------------------------------------------------
    def _tab4view_preferences(self, request, form):
        """Generate the preferences tab.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :type  form: .lib.form.Form
        :param form:
            Current form object.
        :rtype: webhelpers2.html.literal
        """
        translate = request.localizer.translate
        html = ''
        html += form.grid_item(
            translate(_('Hide e-mail:')),
            self.email_hidden and translate(_('yes')), clear=True)
        lang = request.session.get('lang', 'en')
        theme = request.registry['themes'].get(self.theme)
        theme = theme['name'].get(lang, self.theme) if theme else ''
        html += form.grid_item(translate(_('Theme:')), theme, clear=True)
        if self.page_size:
            html += form.grid_item(
                translate(_('Lines per page:')), str(self.page_size),
                clear=True)

        return html or _('No preferences.')

    # -------------------------------------------------------------------------
    def _tab4view_profiles(self, request):
        """Generate the profiles tab.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :rtype: webhelpers2.html.literal
        """
        translate = request.localizer.translate
        if not self.profiles:
            return translate(_('No attributed profile.'))

        html = '<table>\n<thead>\n'\
            '<tr><th>{label}</th><th>{description}</th></tr>\n'\
            '</thead>\n<tbody>\n'.format(
                label=translate(_('Label')),
                description=translate(_('Description')))
        for dbprofile in sorted(self.profiles, key=lambda k: k.profile_id):
            html += \
                '<tr><th><a href="{profile_view}">{label}</a></th>'\
                '<td>{description}</td></tr>\n'.format(
                    profile_view=request.route_path(
                        'profile_view', profile_id=dbprofile.profile_id),
                    label=dbprofile.label(request),
                    description=dbprofile.description(request))
        html += '</tbody>\n</table>\n'

        return literal(html)

    # -------------------------------------------------------------------------
    def _tab4view_groups(self, request):
        """Generate the groups tab.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :rtype: webhelpers2.html.literal
        """
        translate = request.localizer.translate
        if not self.groups:
            return translate(_('This user is in any group.'))

        html = '<table>\n<thead>\n'\
            '<tr><th>{label}</th><th>{description}</th></tr>\n'\
            '</thead>\n<tbody>\n'.format(
                label=translate(_('Label')),
                description=translate(_('Description')))
        for dbgroup in sorted(self.groups, key=lambda k: k.group_id):
            html += \
                '<tr><th><a href="{group_view}">{label}</a></th>'\
                '<td>{description}</td></tr>\n'.format(
                    group_view=request.route_path(
                        'group_view', group_id=dbgroup.group_id),
                    label=dbgroup.label(request),
                    description=dbgroup.description(request))
        html += '</tbody>\n</table>\n'
        return literal(html)

    # -------------------------------------------------------------------------
    @classmethod
    def settings_schema(cls, request, profiles, groups, dbuser=None):
        """Return a Colander schema to edit user account.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :param dict profiles:
            A dictionary such as ``{profile_id: (label, description),...}``.
        :param dict groups:
            A dictionary such as ``{group_id: (label, description),...}``.
        :type  dbuser: DBUser
        :param dbuser: (optional)
            Current user SqlAlchemy object.
        :rtype: tuple
        :return:
            A tuple such as ``(schema, defaults)``.
        """
        # Identity
        schema = colander.SchemaNode(colander.Mapping())

        schema.add(colander.SchemaNode(
            colander.String(), name='login',
            validator=colander.All(
                colander.Regex(r'^[\w@_.-]+$'),
                colander.Length(min=2, max=EMAIL_LEN))))

        schema.add(colander.SchemaNode(
            colander.String(), name='honorific',
            validator=colander.OneOf(
                cls.honorific.property.columns[0].type.enums), missing=None))
        schema.add(colander.SchemaNode(
            colander.String(), name='first_name',
            validator=colander.Length(max=NAME_LEN), missing=None))
        schema.add(colander.SchemaNode(
            colander.String(), name='last_name',
            validator=colander.Length(max=NAME_LEN)))
        schema.add(colander.SchemaNode(
            colander.String(), name='email',
            validator=colander.All(
                colander.Email(), colander.Length(max=LABEL_LEN))))
        schema.add(colander.SchemaNode(
            colander.String(), name='language',
            validator=colander.OneOf(settings_get_list(
                request.registry.settings, 'languages', ['en'])),
            missing=None))
        schema.add(colander.SchemaNode(
            colander.String(), name='time_zone',
            validator=colander.OneOf(common_timezones),
            missing=None))

        # Security
        i_admin = request.has_permission('system.administrator')
        if 'authorities' in request.registry and i_admin:
            schema.add(colander.SchemaNode(
                colander.String(), name='authority',
                validator=colander.All(
                    colander.Regex(r'^[0-9a-z-]+$'),
                    colander.Length(max=ID_LEN)),
                missing=None))
        if dbuser is None or (
                dbuser.user_id != request.session['user']['user_id'] and
                (dbuser.status != 'administrator' or i_admin) and
                request.has_permission('user-edit')):
            status = list(USER_STATUS_LABELS.keys())
            if not i_admin:
                del status[status.index('administrator')]
            schema.add(colander.SchemaNode(
                colander.String(), name='status',
                validator=colander.OneOf(status)))
        password_min_length = request.registry['settings'][
            'password-min-length']
        if dbuser is None:
            schema.add(colander.SchemaNode(
                colander.String(), name='password1',
                validator=colander.Length(min=password_min_length)))
            schema.add(colander.SchemaNode(
                colander.String(), name='password2',
                validator=SameAs(request, 'password1', _(
                    'The two passwords are not identical.'))))
        else:
            schema.add(colander.SchemaNode(
                colander.String(), name='password1', missing=None,
                validator=colander.Length(min=password_min_length)))
            schema.add(colander.SchemaNode(
                colander.String(), name='password2', missing=None,
                validator=SameAs(request, 'password1', _(
                    'The two passwords are not identical.'))))
        schema.add(colander.SchemaNode(
            colander.Boolean(), name='password_mustchange', missing=False))

        # Preferences
        schema.add(colander.SchemaNode(
            colander.Boolean(), name='email_hidden', missing=False))
        if len(request.registry['themes']) > 1:
            schema.add(colander.SchemaNode(
                colander.String(), name='theme',
                validator=colander.OneOf(request.registry['themes']),
                missing=None))
        schema.add(colander.SchemaNode(
            colander.Integer(), name='page_size',
            validator=colander.OneOf(PAGE_SIZES[1:-1]), missing=None))

        if request.has_permission('user-create'):
            cls._settings_schema_creator(profiles, groups, schema)

        # Defaults
        if dbuser is None:
            defaults = {'status': 'active'}
        else:
            defaults = {}
            for dbprofile in dbuser.profiles:
                defaults['pfl:{0}'.format(dbprofile.profile_id)] = True
            for dbgroup in dbuser.groups:
                defaults['grp:{0}'.format(dbgroup.group_id)] = True

        return schema, defaults

    # -------------------------------------------------------------------------
    @classmethod
    def _settings_schema_creator(cls, profiles, groups, schema):
        """Complete the Colander schema if I am a creator.

        :param dict profiles:
            A dictionary such as ``{profile_id: (label, description),...}``.
        :param dict groups:
            A dictionary such as ``{group_id: (label, description),...}``.
        :type  schema: colander.SchemaNode
        :param schema:
            Current schema to complete.
        """
        # Expiration
        schema.add(colander.SchemaNode(
            colander.Date(), name='expiration', missing=None))

        # Profiles
        for profile_id in profiles:
            schema.add(colander.SchemaNode(
                colander.Boolean(), name='pfl:{0}'.format(profile_id),
                missing=False))

        # Groups
        for group_id in groups:
            schema.add(colander.SchemaNode(
                colander.Boolean(), name='grp:{0}'.format(group_id),
                missing=False))

    # -------------------------------------------------------------------------
    @classmethod
    def tab4edit(cls, request, tab_index, form, profiles, groups, dbuser=None):
        """Generate the tab content of user account for edition.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :param int tab_index:
            Index of the tab.
        :type  form: .lib.form.Form
        :param form:
            Current form object.
        :param dict profiles:
            A dictionary such as ``{profile_id: (label, description),...}``.
        :param dict groups:
            A dictionary such as ``{group_id: (label, description),...}``.
        :type  dbuser: DBUser
        :param dbuser: (optional)
            Current user SqlAlchemy object.
        :rtype: webhelpers2.html.literal
        """
        # pylint: disable = too-many-arguments
        if tab_index == 0:
            return cls._tab4edit_information(request, form, dbuser)
        if tab_index == 1:
            return cls._tab4edit_preferences(request, form)
        if tab_index == 2:
            return cls._tab4edit_profiles(request, form, profiles)
        if tab_index == 3:
            return cls._tab4edit_groups(request, form, groups)
        return ''

    # -------------------------------------------------------------------------
    @classmethod
    def _tab4edit_information(cls, request, form, dbuser):
        """Generate the information tab for edition.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :type  form: .lib.form.Form
        :param form:
            Current form object.
        :type  dbuser: DBUser
        :param dbuser:
            Current user SqlAlchemy object.
        :rtype: webhelpers2.html.literal
        """
        # Identify
        i_admin = request.has_permission('system.administrator')
        has_authority = dbuser is not None and dbuser.authority is not None
        translate = request.localizer.translate
        html = HTML.h3(translate(_('Identity')))
        if has_authority and not i_admin:
            html += form.grid_item(
                translate(_('Login:')), dbuser.login, clear=True)
            html += form.hidden('login', dbuser.login)
        else:
            html += form.grid_text(
                'login', translate(_('Login:')), maxlength=EMAIL_LEN,
                required=True, clear=True)
        html += form.grid_select(
            'honorific', translate(_('Title:')),
            [('', ' ')] + list(cls.honorific_labels.items()), clear=True)
        html += form.grid_text(
            'first_name', translate(_('First Name:')), maxlength=NAME_LEN,
            clear=True)
        html += form.grid_text(
            'last_name', translate(_('Last Name:')), maxlength=NAME_LEN,
            required=True, clear=True)
        html += form.grid_text(
            'email', translate(_('Email:')), maxlength=EMAIL_LEN,
            required=True, clear=True)
        html += form.grid_select(
            'language', translate(_('Language:')),
            [('', ' ')] + sorted(settings_get_list(
                request.registry.settings, 'languages', ['en'])), clear=True)
        html += form.grid_select(
            'time_zone', translate(_('Time Zone:')),
            [('', ' ')] + list(common_timezones), clear=True)

        # Security
        html += HTML.h3(translate(_('Security')))
        if 'authorities' in request.registry and \
           request.registry['authorities']:
            if i_admin:
                html += form.grid_select(
                    'authority', translate(_('Authority:')),
                    [('', ' ')] + sorted(request.registry['authorities']),
                    clear=True)
            elif has_authority:
                html += form.grid_item(
                    translate(_('Authority:')), dbuser.authority, clear=True)
        if dbuser is None or (
                dbuser.user_id != request.session['user']['user_id'] and
                (dbuser.status != 'administrator' or i_admin) and
                request.has_permission('user-edit')):
            status = cls.status_labels.items() if i_admin else \
                [k for k in cls.status_labels.items()
                 if k[0] != 'administrator']
            html += form.grid_select(
                'status', translate(_('Status:')), status, clear=True)
        else:
            html += form.grid_item(
                translate(_('Status:')),
                translate(dbuser.status_labels[dbuser.status]), clear=True)
        if not has_authority or i_admin:
            html += form.grid_password(
                'password1', translate(_('Password:')), maxlength=64,
                required=dbuser is None, clear=True)
            html += form.grid_password(
                'password2', translate(_('Confirmation:')), maxlength=64,
                required=dbuser is None, clear=True)
            html += form.grid_custom_checkbox(
                'password_mustchange', translate(_('Force change:')),
                clear=True)
        if request.has_permission('user-create'):
            html += form.grid_text(
                'expiration', translate(_('Expiration:')), maxlength=10,
                hint=translate(_('Format: YYYY-MM-DD')),
                class_='cioFormItem cioDate', clear=True)

        return html

    # -------------------------------------------------------------------------
    @classmethod
    def _tab4edit_preferences(cls, request, form):
        """Generate the preferences tab for edition.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :type  form: .lib.form.Form
        :param form:
            Current form object.
        :rtype: webhelpers2.html.literal
        """
        translate = request.localizer.translate
        html = form.grid_custom_checkbox(
            'email_hidden', translate(_('Hide e-mail:')), clear=True)
        if len(request.registry['themes']) > 1:
            lang = request.session.get('lang', 'en')
            themes = [(i, k['name'].get(lang, i))
                      for i, k in request.registry['themes'].items()]
            html += form.grid_select(
                'theme', translate(_('Theme:')),
                [('', ' ')] + sorted(themes, key=lambda k: k[1]), clear=True)
        html += form.grid_select(
            'page_size', translate(_('Lines per page:')), PAGE_SIZES[:-1],
            clear=True)

        return html

    # -------------------------------------------------------------------------
    @classmethod
    def _tab4edit_profiles(cls, request, form, profiles):
        """Generate the profiles tab for edition.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :type  form: .lib.form.Form
        :param form:
            Current form object.
        :param dict profiles:
            A dictionary such as ``{profile_id: (label, description),...}``.
        :rtype: webhelpers2.html.literal
        """
        translate = request.localizer.translate
        if not profiles:
            return translate(_('No available profile.'))

        if not request.has_permission('user-create'):
            return translate(
                _('You do not have the rigths to edit profiles.'))

        html = '<table>\n<thead>\n'\
            '<tr><th></th><th>{label}</th>'\
            '<th>{description}</th></tr>\n</thead>\n<tbody>\n'.format(
                label=translate(_('Label')),
                description=translate(_('Description')))
        for profile_id in sorted(profiles):
            html += \
                '<tr><td>{selected}</td>'\
                '<th><label for="{id}">{label}</label></th>'\
                '<td>{description}</td></tr>\n'.format(
                    selected=form.custom_checkbox(
                        'pfl:{0}'.format(profile_id)),
                    id='pfl{0}'.format(profile_id),
                    label=profiles[profile_id][0],
                    description=profiles[profile_id][1])
        html += '</tbody>\n</table>\n'

        return literal(html)

    # -------------------------------------------------------------------------
    @classmethod
    def _tab4edit_groups(cls, request, form, groups):
        """Generate the groups tab for edition.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :type  form: .lib.form.Form
        :param form:
            Current form object.
        :param dict groups:
            A dictionary such as ``{group_id: (label, description),...}``.
        :rtype: webhelpers2.html.literal
        """
        translate = request.localizer.translate
        if not groups:
            return translate(_('No available group.'))
        if not request.has_permission('user-create'):
            return translate(
                _('You do not have the rigths to edit groups.'))

        html = '<table>\n<thead>\n'\
            '<tr><th></th><th>{label}</th>'\
            '<th>{description}</th></tr>\n</thead>\n<tbody>\n'.format(
                label=translate(_('Label')),
                description=translate(_('Description')))
        for group_id in sorted(groups):
            html += \
                '<tr><td>{selected}</td>'\
                '<th><label for="{id}">{label}</label></th>'\
                '<td>{description}</td></tr>\n'.format(
                    selected=form.custom_checkbox('grp:{0}'.format(group_id)),
                    id='grp{0}'.format(group_id),
                    label=groups[group_id][0],
                    description=groups[group_id][1])
        html += '</tbody>\n</table>\n'

        return literal(html)


# =============================================================================
class DBUserProfile(DBDeclarativeClass):
    """Class to link users with their profiles (many-to-many)."""
    # pylint: disable = too-few-public-methods

    __tablename__ = 'users_profiles'
    __table_args__ = {'mysql_engine': 'InnoDB'}

    user_id = Column(
        Integer, ForeignKey('users.user_id', ondelete='CASCADE'),
        primary_key=True)
    profile_id = Column(
        String(ID_LEN), ForeignKey('profiles.profile_id', ondelete='CASCADE'),
        primary_key=True)



"""SQLAlchemy-powered model definitions for settings."""

from os.path import basename
from datetime import datetime

from sqlalchemy import Column, String
from sqlalchemy.exc import OperationalError, ProgrammingError
from lxml import etree
from webhelpers2.html import HTML, literal
import colander

from ..lib.i18n import _
from ..lib.utils import tounicode, deltatime_label, size_label
from ..lib.config import settings_get_list
from ..lib.paging import PAGE_DEFAULT_SIZE, PAGE_SIZES
from . import DBDeclarativeClass, ID_LEN, EMAIL_LEN, LABEL_LEN
from .dbbase import DBBaseClass

SETTINGS_DEFAULTS = {
    'language': 'en', 'password-min-length': 8, 'remember-me': 5184000,
    'page-size': PAGE_DEFAULT_SIZE, 'download-max-size': 10485760,
    'clipboard-size': 10, 'theme': ''}


# =============================================================================
class DBSettings(DBDeclarativeClass, DBBaseClass):
    """SQLAlchemy-powered settings class."""

    suffix = 'cioset'
    settings_defaults = SETTINGS_DEFAULTS
    _settings_tabs = (_('Parameters'), _('Information'))

    __tablename__ = 'settings'
    __table_args__ = {'mysql_engine': 'InnoDB'}

    key = Column(String(ID_LEN), primary_key=True)
    value = Column(String(128))

    # -------------------------------------------------------------------------
    @classmethod
    def exists(cls, dbsession):
        """Check if table ``settings`` exists and has at least one item.

        :type  dbsession: sqlalchemy.orm.session.Session
        :param dbsession:
            SQLAlchemy session (for testing).
        :rtype: bool
        """
        try:
            setting = dbsession.query(cls).first()
        except (OperationalError, ProgrammingError, UnicodeEncodeError):
            return False
        return setting is not None

    # -------------------------------------------------------------------------
    @classmethod
    def xml2db(cls, dbsession, settings_elt):
        """Load settings from a XML element.

        :type  dbsession: sqlalchemy.orm.session.Session
        :param dbsession:
            SQLAlchemy session.
        :type  settings_elt: lxml.etree.Element
        :param settings_elt:
            Profile XML element.
        """
        if settings_elt is None:
            return

        # Clean up old settings
        for dbsetting in dbsession.query(cls):
            dbsession.delete(dbsetting)

        # Fill settings table
        dbsession.add(cls(
            key='populate', value=datetime.now().isoformat(' ').split('.')[0]))
        for elt in settings_elt.iterchildren(tag=etree.Element):
            dbsession.add(cls(
                key=elt.tag,
                value=elt.text[:128] if elt.text is not None else None))

    # -------------------------------------------------------------------------
    @classmethod
    def db2xml(cls, dbsession):
        """Serialize settings to a XML representation.

        :type  dbsession: sqlalchemy.orm.session.Session
        :param dbsession:
            SQLAlchemy session.
        :rtype: lxml.etree.Element
        """
        settings_elt = etree.Element('settings')

        # Key/value from database
        for dbsetting in dbsession.query(cls):
            if dbsetting.key != 'populate':
                etree.SubElement(settings_elt, dbsetting.key).text = \
                    dbsetting.value

        return settings_elt if len(settings_elt) + 1 > 1 else None

    # -------------------------------------------------------------------------
    @classmethod
    def db2dict(cls, settings, dbsession, default_email):
        """Return a dictionary containing settings.

        :param dict settings:
            Pyramid settings.
        :type  dbsession: sqlalchemy.orm.session.Session
        :param dbsession:
            SQLAlchemy session (for testing).
        :param str default_email:
            Default email for the site, usually the administrator email.
        :rtype: dict
        """
        try:
            dbsettings = dbsession.query(cls).all()
        except (OperationalError, ProgrammingError, UnicodeEncodeError):
            dbsettings = []

        # Default settings
        settings_dict = cls.settings_defaults.copy()
        if len(dbsettings) < 2:
            settings_dict['title'] = tounicode(settings.get(
                'site.title', settings['site.uid']))
            settings_dict['email'] = default_email
            if 'pyramid.default_locale_name' in settings:
                settings_dict['language'] = settings[
                    'pyramid.default_locale_name']
            settings_dict['populate'] = datetime.now().isoformat(
                ' ').split('.')[0]
            return settings_dict

        # ... or customized settings
        for dbsetting in dbsettings:
            settings_dict[dbsetting.key] = dbsetting.value
        settings_dict['password-min-length'] = int(
            settings_dict['password-min-length'])
        settings_dict['remember-me'] = int(settings_dict['remember-me'])
        settings_dict['page-size'] = int(settings_dict['page-size'])
        settings_dict['download-max-size'] = int(
            settings_dict['download-max-size'])
        settings_dict['clipboard-size'] = int(settings_dict['clipboard-size'])
        if 'chrysalio.includes' not in settings or \
           'chrysalio.includes.themes' not in settings['chrysalio.includes']:
            settings_dict['theme'] = ''

        return settings_dict

    # -------------------------------------------------------------------------
    @classmethod
    def tab4view(cls, request, tab_index, form):
        """Generate the tab content for settings.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :param int index:
            Index of the tab.
        :type  form: .lib.form.Form
        :param tuple principals:
            Groups of principals.
        :rtype: webhelpers2.html.literal
        """
        if tab_index == 0:
            return cls._tab4view_parameters(request, form)
        if tab_index == 1:
            return cls._tab4view_information(request, form)
        return ''

    # -------------------------------------------------------------------------
    @classmethod
    def _tab4view_parameters(cls, request, form):
        """Generate the parameters tab.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :type  form: .lib.form.Form
        :param form:
            Current form object.
        :rtype: webhelpers2.html.literal
        """
        translate = request.localizer.translate
        html = HTML.h3(translate(_('Site')))
        html += form.grid_item(
            translate(_('Title:')), request.registry['settings']['title'],
            clear=True)
        html += form.grid_item(
            translate(_('Email:')),
            request.registry['settings'].get('email'), clear=True)

        html += HTML.h3(translate(_('Default values')))
        html += form.grid_item(
            translate(_('Language:')),
            request.registry['settings']['language'], clear=True)
        html += form.grid_item(
            translate(_('Password length:')),
            request.registry['settings']['password-min-length'], clear=True)
        html += form.grid_item(
            translate(_('Remember me:')),
            deltatime_label(
                request.registry['settings']['remember-me'],
                lang=request.locale_name),
            clear=True)
        html += form.grid_item(
            translate(_('Lines per page:')),
            request.registry['settings']['page-size'], clear=True)
        if request.registry['settings']['download-max-size']:
            html += form.grid_item(
                translate(_('Maximum size of download:')),
                size_label(request.registry['settings']['download-max-size']),
                clear=True)
        html += form.grid_item(
            translate(_('Clipboard size:')),
            request.registry['settings']['clipboard-size'], clear=True)
        lang = request.session.get('lang', 'en')
        html += form.grid_item(
            translate(_('Theme:')), request.registry['themes'][
                request.registry['settings']['theme']]['name'].get(
                    lang, request.registry['settings']['theme']),
            clear=True)

        return html

    # -------------------------------------------------------------------------
    @classmethod
    def _tab4view_information(cls, request, form, readonly_message=False):
        """Generate the information tab.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :type  form: .lib.form.Form
        :param form:
            Current form object.
        :param bool readonly_message:
            If ``True`` add a `Read Only` message.
        :rtype: webhelpers2.html.literal
        """
        translate = request.localizer.translate

        html = ''
        if readonly_message:
            html = HTML.div(
                translate(_('You cannot modify these values.')),
                class_='cioAlert')

        html += HTML.h3(translate(_('Site')))
        html += form.grid_item(
            translate(_('Identifier:')),
            request.registry.settings['site.uid'], clear=True)
        html += form.grid_item(
            translate(_('Application version:')), request.registry['version'],
            clear=True)
        html += form.grid_item(
            translate(_('Configuration version:')),
            request.registry.settings.get('site.version'), clear=True)

        html += HTML.h3(translate(_('Database')))
        html += form.grid_item(
            translate(_('Name:')),
            basename(request.dbsession.bind.url.database), clear=True)
        html += form.grid_item(
            translate(_('Type:')), request.dbsession.bind.name, clear=True)
        html += form.grid_item(
            translate(_('Last loading:')),
            request.registry['settings']['populate'], clear=True)

        html += HTML.h3(translate(_('Resources')))
        html += form.grid_item(
            translate(_('Log:')),
            translate(_('active')) if request.registry.get('log_activity')
            else translate(_('inactive')), clear=True)
        html += form.grid_item(
            translate(_('Languages:')), ', '.join(settings_get_list(
                request.registry.settings, 'languages', ['en'])),
            clear=True)
        lang = request.session.get('lang', 'en')
        themes = sorted([
            request.registry['themes'][k]['name'].get(lang, k)
            for k in request.registry['themes']])
        html += form.grid_item(
            translate(_('Themes:')), ', '.join(themes), clear=True)

        return html

    # -------------------------------------------------------------------------
    @classmethod
    def settings_schema(cls, request):
        """Return a Colander schema to edit settings.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :rtype: tuple
        :return:
            A tuple such as ``(schema, defaults)``.
        """
        # Site
        schema = colander.SchemaNode(colander.Mapping())
        schema.add(colander.SchemaNode(
            colander.String(), name='title',
            validator=colander.Length(max=LABEL_LEN)))
        schema.add(colander.SchemaNode(
            colander.String(), name='email',
            validator=colander.All(
                colander.Length(max=EMAIL_LEN), colander.Email())))

        # Default values
        schema.add(colander.SchemaNode(
            colander.String(), name='language',
            validator=colander.OneOf(settings_get_list(
                request.registry.settings, 'languages', ['en']))))
        schema.add(colander.SchemaNode(
            colander.Integer(), name='password-min-length',
            validator=colander.Range(min=4)))
        schema.add(colander.SchemaNode(
            colander.Integer(), name='remember-me',
            validator=colander.Range(min=3600)))
        schema.add(colander.SchemaNode(
            colander.Integer(), name='page-size',
            validator=colander.OneOf(PAGE_SIZES[1:-1])))
        schema.add(colander.SchemaNode(
            colander.Integer(), name='download-max-size',
            validator=colander.Range(min=0)))
        schema.add(colander.SchemaNode(
            colander.Integer(), name='clipboard-size',
            validator=colander.Range(min=1)))
        if len(request.registry['themes']) > 1:
            schema.add(colander.SchemaNode(
                colander.String(), name='theme',
                validator=colander.OneOf(request.registry['themes'])))

        return schema, request.registry['settings']

    # -------------------------------------------------------------------------
    @classmethod
    def tab4edit(cls, request, tab_index, form):
        """Generate the tab content for edition.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :param int tab_index:
            Index of the tab.
        :type  form: .lib.form.Form
        :param form:
            Current form object.
        :rtype: webhelpers2.html.literal
        """
        if tab_index == 0:
            return cls._tab4edit_parameters(request, form)
        if tab_index == 1:
            return cls._tab4view_information(request, form, True)
        return ''

    # -------------------------------------------------------------------------
    @classmethod
    def _tab4edit_parameters(cls, request, form):
        """Generate the parameters tab for edition.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :type  form: .lib.form.Form
        :param form:
            Current form object.
        :rtype: webhelpers2.html.literal
        """
        translate = request.localizer.translate
        html = HTML.h3(translate(_('Site')))
        html += form.grid_text(
            'title', translate(_('Title:')), maxlength=LABEL_LEN,
            required=True, clear=True)
        html += form.grid_text(
            'email', translate(_('Email:')), maxlength=EMAIL_LEN,
            required=True, clear=True)

        html += HTML.h3(translate(_('Default values')))
        html += form.grid_select(
            'language', translate(_('Language:')),
            sorted(settings_get_list(
                request.registry.settings, 'languages', ['en'])),
            required=True, clear=True)
        html += form.grid_text(
            'password-min-length', translate(_('Password length:')),
            maxlength=2, required=True, clear=True)
        html += form.grid_text(
            'remember-me', translate(_('Remember me:')),
            maxlength=8, required=True, clear=True)
        html += form.grid_select(
            'page-size', translate(_('Lines per page:')), PAGE_SIZES[:-1],
            required=True, clear=True)
        html += form.grid_text(
            'download-max-size', translate(_('Maximum size of download:')),
            maxlength=10, required=True, clear=True)
        html += form.grid_text(
            'clipboard-size', translate(_('Clipboard size:')),
            maxlength=2, required=True, clear=True)
        if len(request.registry['themes']) > 1:
            lang = request.session.get('lang', 'en')
            themes = [(i, j['name'].get(lang, i))
                      for i, j in request.registry['themes'].items()]
            html += form.grid_select(
                'theme', translate(_('Theme:')),
                sorted(themes, key=lambda k: k[1]), clear=True)

        return literal(html)



# pylint: disable = import-outside-toplevel
"""Tests of ``menu`` functions."""

from unittest import TestCase

from . import ConfiguratorTestCase


# =============================================================================
class UMenuIncludeme(ConfiguratorTestCase):
    """Unit test class for :func:`menu.includeme`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:menu.includeme]"""
        self.configurator.include('..menu')
        self.assertIn('menu', self.configurator.registry)


# =============================================================================
class UMenuBeforeRender(TestCase):
    """Unit test class for :func:`menu.before_render`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:menu.before_render]"""
        from pyramid.testing import DummyRequest
        from ..lib.menu import Menu
        from ..menu import before_render

        request = DummyRequest()
        request.registry['menu'] = ''
        event = {'request': request}
        before_render(event)
        self.assertIn('menu', event)
        self.assertIsInstance(event['menu'], Menu)



"""FTP library."""

from sys import version_info
from os import makedirs
from os.path import join, exists, getsize, splitext, basename
from ftplib import FTP, FTP_TLS, all_errors

from .i18n import _
from .utils import EXCLUDED_FILES, tounicode, tostr


# =============================================================================
class Ftp(object):
    """Class to manage FTP.

    :param log_error:
        Function to record errors.
    """

    # -------------------------------------------------------------------------
    def __init__(self, log_error):
        """Constructor method."""
        self._log_error = log_error
        self.connection = None

    # -------------------------------------------------------------------------
    def connect(self, values):
        """FTP connection using ``values`` to get FTP information.

        :param dict values:
            FTP values.
        :rtype: bool
        """
        # Previous connection?
        self.quit()

        # Check FTP arguments
        host = values.get('ftp_host')
        user = values.get('ftp_user', 'anonymous')
        password = values.get('ftp_password', 'anonymous@')
        if host is None:
            self._log_error(_('FTP error: bad arguments!'))
            return False

        # Connection
        self.connection = FTP_TLS() if values.get('ftp_tls') else FTP()
        try:
            self.connection.connect(host, values.get('ftp_port', 21))
            self.connection.login(user, password)
        except all_errors as error:
            self.connection = None
            self._log_ftp_error(error)
            return False
        self.connection.set_pasv(values.get('ftp_pasv', False))

        # Change to root directory
        if values.get('ftp_path') and not self.cwd(values['ftp_path']):
            self.quit()
            return False

        return True

    # -------------------------------------------------------------------------
    def cwd(self, directory):
        """Change working directory.

        :param str directory:
            Directory.
        :rtype: bool
        """
        try:
            self.connection.cwd(directory)
        except all_errors as error:
            self._log_ftp_error(error)
            return False
        return True

    # -------------------------------------------------------------------------
    def mkdir(self, directory):
        """Make a directory.

        :param str directory:
            Directory.
        :rtype: bool
        """
        try:
            self.connection.mkd(directory)
        except all_errors as error:  # pragma: nocover
            self._log_ftp_error(error)
            return False
        return True

    # -------------------------------------------------------------------------
    def list_directory(self):
        """Return the list of files in the current FTP directory.

        :rtype: tuple
        :return:
           A tuple such as ``(dirs_infos, files_infos)``.
        """
        listing = []
        try:
            self.connection.retrlines('MLSD', listing.append)
        except all_errors as error:  # pragma: nocover
            self._log_ftp_error(error)
            return {}, {}

        dirs = {}
        files = {}
        for item in listing:
            item = item.partition(' ')
            infos = {
                k.split('=')[0].lower():  k.split('=')[1]
                for k in item[0].split(';') if k}
            name = item[2]
            if infos['type'] == 'dir':
                dirs[name] = infos
            elif infos['type'] == 'file':
                files[name] = infos

        return dirs, files

    # -------------------------------------------------------------------------
    def delete(self, filename):
        """Delete the file ``filename``.

        :rtype: bool
        """
        try:
            self.connection.delete(filename)
        except all_errors as error:  # pragma: nocover
            self._log_ftp_error(error)
            return False
        return True

    # -------------------------------------------------------------------------
    def rmtree(self, directory):
        """Remove the directory ``directory``.

        :rtype: bool
        """
        if not self.cwd(directory):
            return False

        dirs, files = self.list_directory()
        for name in files:
            if not self.delete(name):
                return False  # pragma: nocover
        for name in dirs:
            if not self.rmtree(name):
                return False  # pragma: nocover

        self.connection.cwd('..')
        try:
            self.connection.rmd(directory)
        except all_errors as error:  # pragma: nocover
            self._log_ftp_error(error)
            return False

        return True

    # -------------------------------------------------------------------------
    def download(self, destination, exclude=None):
        """Transfer the current FTP directory into the ``destination``
        directory.

        :rtype: bool
        :return:
            ``True`` if the download completed.
        :param list exclude: (optional)
            List of files to exclude.
        """
        # pylint: disable = too-many-branches
        dirs, files = self.list_directory()
        if not dirs and not files:
            return True

        if not exists(destination):
            makedirs(destination)
        if exclude is None:
            exclude = EXCLUDED_FILES

        completed = True
        for name, infos in files.items():
            if name in exclude:
                continue
            if splitext(name)[1] in ('.part', '.filepart'):
                completed = False
                continue
            try:
                fullname = join(destination, tounicode(name))
            except UnicodeDecodeError as error:  # pragma: nocover
                self._log_ftp_error(error)
                continue
            if not exists(fullname) or \
               getsize(fullname) != int(infos['size']):
                with open(fullname, 'wb') as hdl:
                    self.connection.retrbinary(
                        tostr('RETR {0}'.format(tounicode(name))),
                        hdl.write)
                completed = False

        for name in dirs:
            if name in exclude:
                continue
            try:
                self.connection.cwd(name)
            except all_errors as error:  # pragma: nocover
                self._log_ftp_error(error)
                continue
            try:
                new_destination = join(destination, tounicode(name))
            except UnicodeDecodeError as error:  # pragma: nocover
                self._log_ftp_error(error)
                continue
            if not exists(new_destination):
                makedirs(new_destination)
            completed &= self.download(new_destination)
            self.connection.cwd('..')

        return completed

    # -------------------------------------------------------------------------
    def upload(self, filename, upload_name=None):
        """Upload a file in the current directory.

        :param str filename:
            Absolute path to the file to upload.
        :param str upload_name: (optional)
            Name of the uploaded file.
        """
        stor = 'STOR {0}'.format(upload_name or basename(filename))
        stor = stor.encode('utf8') if version_info < (3, 0) else stor
        with open(filename, 'rb') as hdl:
            self.connection.storbinary(stor, hdl)

    # -------------------------------------------------------------------------
    def quit(self):
        """Quit the FTP connection."""
        if self.connection is not None:
            try:
                self.connection.quit()
                self.connection.close()
            except all_errors as error:  # pragma: nocover
                self._log_ftp_error(error)
            self.connection = None

    # -------------------------------------------------------------------------
    def _log_ftp_error(self, error):
        """Add an error message in the log cache.

        :param str text:
            Error.
        """
        self._log_error(_('FTP error: ${e}', {'e': tounicode(error)}))



"""View callables."""

from pyramid.view import view_config
from pyramid.httpexceptions import HTTPFound

from ...lib.i18n import _
from ...lib.log import log_info
from ...lib.form import get_action, Form
from ...lib.menu import Menu
from ...lib.modes import Modes
from ...views import BaseView
from .models import DBModule


# =============================================================================
class ModulesView(BaseView):
    """Class to manage modules.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    """

    # -------------------------------------------------------------------------
    @view_config(
        route_name='modules_view',
        renderer='chrysalio:includes/modules/Templates/modules_view.pt',
        permission='modules-view')
    def view(self):
        """Show modules."""
        i_editor = self._request.has_permission('modules-edit')

        self._request.breadcrumbs(
            _('Modules'), 2,
            replace=self._request.route_path('modules_edit'))

        return {
            'form': Form(self._request), 'dbmodules': DBModule,
            'i_editor': i_editor}

    # -------------------------------------------------------------------------
    @view_config(
        route_name='modules_edit',
        renderer='chrysalio:includes/modules/Templates/modules_edit.pt',
        permission='modules-edit')
    def edit(self):
        """Edit modules."""
        form = Form(self._request, *DBModule.settings_schema(self._request))
        action = get_action(self._request)[0]
        if action == 'sav!' and form.validate():
            self._save([k for k in form.values if form.values[k]])
            log_info(self._request, 'modules_edit')
            return HTTPFound(self._request.route_path('modules_view'))

        self._request.breadcrumbs(
            _('Modules Edition'), 2,
            replace=self._request.route_path('modules_view'))

        return {'form': form, 'dbmodules': DBModule, 'action': action}

    # -------------------------------------------------------------------------
    def _save(self, after):
        """Save modules.

        :param list after:
            List of desired active modules after update.
        """
        modules = self._request.registry['modules']
        before = [k for k in modules
                  if k not in self._request.registry['modules_off']]
        active = set()
        to_activate = set()

        # Process newly activated
        for module_id in modules:
            if module_id in after:
                if module_id in before:
                    active.add(module_id)
                else:
                    to_activate.add(module_id)
                    self._activate_module_dependencies(module_id, to_activate)

        # Process newly deactivated
        for module_id in modules:
            if module_id not in after and module_id in before and \
               module_id not in to_activate:
                self._deactivate_module_depending(module_id, active)

        # Modules to activave
        to_activate |= active

        # Activate/deactivate modules and update modules table
        self._update_modules(modules, to_activate)

        # Update modes and menu
        Modes.invalidate(self._request, 'modes')
        Menu.invalidate(self._request, 'menu')

    # -------------------------------------------------------------------------
    def _update_modules(self, modules, to_activate):
        """Activate/deactivate modules and update modules table.

        :type  modules: OrderedDict
        :param modules:
            Dictionary of available modules.
        :param set to_activate:
            Set of modules to activate.
        """
        before = [k for k in self._request.registry['modules']
                  if k not in self._request.registry['modules_off']]
        for module_id in modules:
            newly_active = module_id in to_activate

            if newly_active and module_id not in before:
                modules[module_id].activate(
                    self._request.registry, self._request.dbsession)
                self._request.registry['modules_off'].remove(module_id)
            elif not newly_active and module_id in before:
                modules[module_id].deactivate(
                    self._request.registry, self._request.dbsession)
                self._request.registry['modules_off'].add(module_id)

            dbmodule = self._request.dbsession.query(DBModule).filter_by(
                module_id=module_id).first()
            if newly_active and dbmodule is not None:
                self._request.dbsession.delete(dbmodule)
            elif not newly_active and dbmodule is None:
                self._request.dbsession.add(DBModule(
                    module_id=module_id, inactive=True))

    # -------------------------------------------------------------------------
    def _activate_module_dependencies(self, module_id, activated):
        """Recursive method to activate dependencies of the module
        ``module_id``.

        :param str module_id:
            ID of the reference module.
        :param set activated:
            Set of already activated moodules
        """
        modules = self._request.registry['modules']
        for depend_id in modules[module_id].dependencies:
            if depend_id in modules and depend_id not in activated:
                activated.add(depend_id)
                self._activate_module_dependencies(depend_id, activated)

    # -------------------------------------------------------------------------
    def _deactivate_module_depending(self, module_id, active):
        """Recursive method to deactivate modules depending on the module
        ``module_id``.

        :param str module_id:
            ID of the reference module.
        :param set activated:
            Set of already activated moodules
        """
        for active_id in active.copy():
            if module_id in self._request.registry[
                    'modules'][active_id].dependencies:
                active.remove(active_id)
                self._deactivate_module_depending(active_id, active)



"""Manager login view callables."""

from time import time
from datetime import date
from hashlib import sha1

from colander import Mapping, SchemaNode, String, Boolean

from pyramid.httpexceptions import HTTPFound
from pyramid.view import view_config
from pyramid.security import forget, remember

from ..lib.form import Form
from ..lib.log import log_info
from ..models.dbuser import DBUser


# =============================================================================
def login(request, dbuser_class=DBUser):
    """This view renders a login form and processes the post checking
    credentials.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    :type  dbuser_class: .models.dbuser.DBuser
    :param dbuser_class: (default=DBUser)
        Class to manage user.
    """
    # Next step
    next_path = request.params.get('next') \
        or (request.url != request.route_url('login') and request.url) \
        or request.route_url('home')

    # Create form
    schema = SchemaNode(Mapping())
    schema.add(SchemaNode(String(), name='login'))
    schema.add(SchemaNode(String(), name='password'))
    schema.add(SchemaNode(Boolean(), name='remember', missing=False))
    form = Form(request, schema=schema, defaults={'next': next_path})

    # Validate form
    if form.validate():
        dbuser, error = dbuser_class.get(request)
        if dbuser is not None and dbuser.password_mustchange:
            token = '{0}{1}'.format(dbuser.login, date.today().isoformat())
            token = sha1(token.encode('utf8')).hexdigest()
            return HTTPFound(request.route_path(
                'user_password_reset', user_id=dbuser.user_id, token=token))
        if dbuser is not None:
            dbuser.set_session(request)
            log_info(request, 'login')
            max_age = request.registry['settings']['remember-me'] \
                if form.values['remember'] else None
            return HTTPFound(location=next_path, headers=remember(
                request, request.session['user']['login'], max_age=max_age))
        request.session.flash(error, 'alert')

    return {'form': form, 'next': next_path}


# =============================================================================
@view_config(route_name='logout')
def logout(request):
    """This view will clear the credentials of the logged in user and redirect
    back to the login page.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    """
    if 'user' in request.session:
        log_info(request, 'logout')
    request.session.clear()
    request.session['_creation_time'] = time()
    request.session['_accessed_time'] = time()
    return HTTPFound(
        location=request.route_path('login'), headers=forget(request))



"""Chrysalio main Relax NG."""

from os.path import join, dirname

from pyramid.config import Configurator


RELAXNG = {
    'root': 'chrysalio', 'version': '1.0',
    'file': join(dirname(__file__), 'RelaxNG', 'chrysalio.rng')}


# =============================================================================
def includeme(configurator):
    """Function to include Relax NG.

    :type  configurator: pyramid.config.Configurator
    :param configurator:
        Object used to do configuration declaration within the application.

    ``configurator.registry['relaxng']`` is a dictionary describing the main
    Relax NG with the following keys:

    * ``'root'``: the name of the root element, possibly with namespace
    * ``'version'``:  the value of attribute version
    * ``'file'``: the path to the Relax NG file
    """
    if isinstance(configurator, Configurator):
        configurator.registry['relaxng'] = RELAXNG
